# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures
import oslo_messaging as messaging
from oslo_utils.fixture import uuidsentinel as uuids

from nova.api.openstack.compute import services
from nova.compute import api as compute
from nova import context
from nova import exception
from nova import objects
from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.objects import test_objects
from nova.tests.unit.objects import test_service


class ComputeHostAPITestCase(test.TestCase):
    def setUp(self):
        super(ComputeHostAPITestCase, self).setUp()
        self.host_api = compute.HostAPI()
        self.aggregate_api = compute.AggregateAPI()
        self.ctxt = context.get_admin_context()
        self.notifier = self.useFixture(
            nova_fixtures.NotificationFixture(self))
        self.req = fakes.HTTPRequest.blank('')
        self.controller = services.ServiceController()
        self.useFixture(nova_fixtures.SingleCellSimple())

    def _compare_obj(self, obj, db_obj):
        test_objects.compare_obj(self, obj, db_obj,
                                 allow_missing=test_service.OPTIONAL)

    def _compare_objs(self, obj_list, db_obj_list):
        self.assertEqual(len(obj_list), len(db_obj_list),
                         "The length of two object lists are different.")
        for index, obj in enumerate(obj_list):
            self._compare_obj(obj, db_obj_list[index])

    def test_set_host_enabled(self):

        @mock.patch.object(self.host_api.rpcapi, 'set_host_enabled',
                           return_value='fake-result')
        @mock.patch.object(self.host_api, '_assert_host_exists',
                           return_value='fake_host')
        def _do_test(mock_assert_host_exists, mock_set_host_enabled):
            result = self.host_api.set_host_enabled(self.ctxt, 'fake_host',
                                                    'fake_enabled')
            self.assertEqual('fake-result', result)
            self.assertEqual(2, len(self.notifier.notifications))
            msg = self.notifier.notifications[0]
            self.assertEqual('HostAPI.set_enabled.start', msg.event_type)
            self.assertEqual('api.fake_host', msg.publisher_id)
            self.assertEqual('INFO', msg.priority)
            self.assertEqual('fake_enabled', msg.payload['enabled'])
            self.assertEqual('fake_host', msg.payload['host_name'])
            msg = self.notifier.notifications[1]
            self.assertEqual('HostAPI.set_enabled.end', msg.event_type)
            self.assertEqual('api.fake_host', msg.publisher_id)
            self.assertEqual('INFO', msg.priority)
            self.assertEqual('fake_enabled', msg.payload['enabled'])
            self.assertEqual('fake_host', msg.payload['host_name'])

        _do_test()

    def test_host_name_from_assert_hosts_exists(self):
        @mock.patch.object(self.host_api.rpcapi, 'set_host_enabled',
                           return_value='fake-result')
        @mock.patch.object(self.host_api, '_assert_host_exists',
                           return_value='fake_host')
        def _do_test(mock_assert_host_exists, mock_set_host_enabled):
            result = self.host_api.set_host_enabled(self.ctxt, 'fake_host',
                                                    'fake_enabled')
            self.assertEqual('fake-result', result)

        _do_test()

    def test_get_host_uptime(self):
        @mock.patch.object(self.host_api.rpcapi, 'get_host_uptime',
                           return_value='fake-result')
        @mock.patch.object(self.host_api, '_assert_host_exists',
                           return_value='fake_host')
        def _do_test(mock_assert_host_exists, mock_get_host_uptime):
            result = self.host_api.get_host_uptime(self.ctxt, 'fake_host')
            self.assertEqual('fake-result', result)

        _do_test()

    @mock.patch('nova.db.main.api.service_get_by_compute_host')
    def test_get_host_uptime_service_down(
        self, mock_get_service_get_by_compute_host,
    ):
        mock_get_service_get_by_compute_host.return_value = dict(
            test_service.fake_service, id=1)

        with mock.patch.object(
            self.host_api.servicegroup_api, 'service_is_up',
            return_value=False,
        ):
            self.assertRaises(exception.ComputeServiceUnavailable,
                              self.host_api.get_host_uptime, self.ctxt,
                              'fake_host')

    def test_host_power_action(self):

        @mock.patch.object(self.host_api.rpcapi, 'host_power_action',
                           return_value='fake-result')
        @mock.patch.object(self.host_api, '_assert_host_exists',
                           return_value='fake_host')
        def _do_test(mock_assert_host_exists, mock_host_power_action):
            result = self.host_api.host_power_action(self.ctxt, 'fake_host',
                                                     'fake_action')
            self.assertEqual('fake-result', result)
            self.assertEqual(2, len(self.notifier.notifications))
            msg = self.notifier.notifications[0]
            self.assertEqual('HostAPI.power_action.start', msg.event_type)
            self.assertEqual('api.fake_host', msg.publisher_id)
            self.assertEqual('INFO', msg.priority)
            self.assertEqual('fake_action', msg.payload['action'])
            self.assertEqual('fake_host', msg.payload['host_name'])
            msg = self.notifier.notifications[1]
            self.assertEqual('HostAPI.power_action.end', msg.event_type)
            self.assertEqual('api.fake_host', msg.publisher_id)
            self.assertEqual('INFO', msg.priority)
            self.assertEqual('fake_action', msg.payload['action'])
            self.assertEqual('fake_host', msg.payload['host_name'])

        _do_test()

    def test_set_host_maintenance(self):

        @mock.patch.object(self.host_api.rpcapi, 'host_maintenance_mode',
                           return_value='fake-result')
        @mock.patch.object(self.host_api, '_assert_host_exists',
                           return_value='fake_host')
        def _do_test(mock_assert_host_exists, mock_host_maintenance_mode):
            result = self.host_api.set_host_maintenance(self.ctxt, 'fake_host',
                                                        'fake_mode')
            self.assertEqual('fake-result', result)
            self.assertEqual(2, len(self.notifier.notifications))
            msg = self.notifier.notifications[0]
            self.assertEqual('HostAPI.set_maintenance.start', msg.event_type)
            self.assertEqual('api.fake_host', msg.publisher_id)
            self.assertEqual('INFO', msg.priority)
            self.assertEqual('fake_host', msg.payload['host_name'])
            self.assertEqual('fake_mode', msg.payload['mode'])
            msg = self.notifier.notifications[1]
            self.assertEqual('HostAPI.set_maintenance.end', msg.event_type)
            self.assertEqual('api.fake_host', msg.publisher_id)
            self.assertEqual('INFO', msg.priority)
            self.assertEqual('fake_host', msg.payload['host_name'])
            self.assertEqual('fake_mode', msg.payload['mode'])

        _do_test()

    def test_service_get_all_cells(self):
        cells = objects.CellMappingList.get_all(self.ctxt)
        for cell in cells:
            with context.target_cell(self.ctxt, cell) as cctxt:
                objects.Service(context=cctxt,
                                binary='nova-compute',
                                host='host-%s' % cell.uuid).create()
        services = self.host_api.service_get_all(self.ctxt, all_cells=True)
        self.assertEqual(sorted(['host-%s' % cell.uuid for cell in cells]),
                         sorted([svc.host for svc in services]))

    @mock.patch('nova.context.scatter_gather_cells')
    def test_service_get_all_cells_with_failures(self, mock_sg):
        service = objects.Service(binary='nova-compute',
                                  host='host-%s' % uuids.cell1)
        mock_sg.return_value = {
            uuids.cell1: [service],
            uuids.cell2: context.did_not_respond_sentinel
        }
        services = self.host_api.service_get_all(self.ctxt, all_cells=True)
        # returns the results from cell1 and ignores cell2.
        self.assertEqual(['host-%s' % uuids.cell1],
                         [svc.host for svc in services])

    @mock.patch('nova.context.scatter_gather_cells')
    def test_service_get_all_cell0_computes(self, mock_sg):
        service = objects.Service(binary='nova-compute', host='rogue')
        mock_sg.return_value = {
            objects.CellMapping.CELL0_UUID: [service],
        }
        with mock.patch.object(compute, 'LOG') as mock_log:
            services = self.host_api.service_get_all(self.ctxt, all_cells=True)
            mock_log.warning.assert_called_once_with(
                'Found compute service %(service)s in cell0; '
                'This should never happen!',
                {'service': 'rogue'})
            self.assertEqual([service], services)

    @mock.patch('nova.objects.CellMappingList.get_all')
    @mock.patch.object(objects.HostMappingList, 'get_by_cell_id')
    @mock.patch('nova.context.scatter_gather_all_cells')
    def test_service_get_all_cells_with_minimal_constructs(self, mock_sg,
                                                           mock_get_hm,
                                                           mock_cm_list):
        service = objects.Service(binary='nova-compute',
                                  host='host-%s' % uuids.cell0)
        cells = [
            objects.CellMapping(uuid=uuids.cell1, id=1),
            objects.CellMapping(uuid=uuids.cell2, id=2),
        ]
        mock_cm_list.return_value = cells
        context.load_cells()
        # create two hms in cell1, which is the down cell in this test.
        hm1 = objects.HostMapping(self.ctxt, host='host1-unavailable',
            cell_mapping=cells[0])
        hm1.create()
        hm2 = objects.HostMapping(self.ctxt, host='host2-unavailable',
            cell_mapping=cells[0])
        hm2.create()
        mock_sg.return_value = {
            cells[0].uuid: [service],
            cells[1].uuid: context.did_not_respond_sentinel,
        }
        mock_get_hm.return_value = [hm1, hm2]
        services = self.host_api.service_get_all(self.ctxt, all_cells=True,
            cell_down_support=True)
        # returns the results from cell0 and minimal construct from cell1.
        self.assertEqual(sorted(['host-%s' % uuids.cell0, 'host1-unavailable',
                         'host2-unavailable']),
                         sorted([svc.host for svc in services]))
        mock_sg.assert_called_once_with(self.ctxt, objects.ServiceList.get_all,
                                        None, set_zones=False)
        mock_get_hm.assert_called_once_with(self.ctxt, cells[1].id)

    @mock.patch('nova.db.main.api.service_get_all')
    def test_service_get_all_no_zones(self, mock_service_get_all):
        services = [dict(test_service.fake_service,
                         id=1, topic='compute', host='host1'),
                    dict(test_service.fake_service,
                         topic='compute', host='host2')]

        mock_service_get_all.return_value = services
        # Test no filters
        result = self.host_api.service_get_all(self.ctxt)
        mock_service_get_all.assert_called_once_with(self.ctxt,
                                                     disabled=None)
        self._compare_objs(result, services)

        # Test no filters #2
        mock_service_get_all.reset_mock()
        result = self.host_api.service_get_all(self.ctxt, filters={})
        mock_service_get_all.assert_called_once_with(self.ctxt,
                                                     disabled=None)
        self._compare_objs(result, services)

        # Test w/ filter
        mock_service_get_all.reset_mock()
        result = self.host_api.service_get_all(self.ctxt,
                                               filters=dict(host='host2'))
        mock_service_get_all.assert_called_once_with(self.ctxt,
                                                     disabled=None)
        self._compare_objs(result, [services[1]])

    @mock.patch('nova.db.main.api.service_get_all')
    def test_service_get_all(self, mock_service_get_all):
        services = [dict(test_service.fake_service,
                         topic='compute', host='host1'),
                    dict(test_service.fake_service,
                         topic='compute', host='host2')]
        exp_services = []
        for service in services:
            exp_service = {}
            exp_service.update(availability_zone='nova', **service)
            exp_services.append(exp_service)

        mock_service_get_all.return_value = services

        # Test no filters
        result = self.host_api.service_get_all(self.ctxt, set_zones=True)
        mock_service_get_all.assert_called_once_with(self.ctxt,
                                                     disabled=None)
        self._compare_objs(result, exp_services)

        # Test no filters #2
        mock_service_get_all.reset_mock()
        result = self.host_api.service_get_all(self.ctxt, filters={},
                                               set_zones=True)
        mock_service_get_all.assert_called_once_with(self.ctxt,
                                                     disabled=None)
        self._compare_objs(result, exp_services)

        # Test w/ filter
        mock_service_get_all.reset_mock()
        result = self.host_api.service_get_all(self.ctxt,
                                               filters=dict(host='host2'),
                                               set_zones=True)
        mock_service_get_all.assert_called_once_with(self.ctxt,
                                                     disabled=None)
        self._compare_objs(result, [exp_services[1]])

        # Test w/ zone filter but no set_zones arg.
        mock_service_get_all.reset_mock()
        filters = {'availability_zone': 'nova'}
        result = self.host_api.service_get_all(self.ctxt,
                                               filters=filters)
        mock_service_get_all.assert_called_once_with(self.ctxt,
                                                     disabled=None)
        self._compare_objs(result, exp_services)

    @mock.patch(
        'nova.db.main.api.service_get_by_compute_host',
        return_value=test_service.fake_service)
    def test_service_get_by_compute_host(
        self, mock_service_get_by_compute_host,
    ):
        result = self.host_api.service_get_by_compute_host(
            self.ctxt, 'fake-host')
        self.assertEqual(test_service.fake_service['id'], result.id)

    @mock.patch('nova.db.main.api.service_get_by_host_and_binary')
    @mock.patch('nova.db.main.api.service_update')
    def test_service_update_by_host_and_binary(
        self, mock_service_update, mock_service_get_by_host_and_binary,
    ):
        host_name = 'fake-host'
        binary = 'nova-compute'
        params_to_update = dict(disabled=True)
        service_id = 42
        expected_result = dict(test_service.fake_service, id=service_id)

        mock_service_get_by_host_and_binary.return_value = expected_result
        mock_service_update.return_value = expected_result

        with mock.patch.object(
            self.host_api, '_update_compute_provider_status',
        ) as mock_update_compute_provider_status:
            result = self.host_api.service_update_by_host_and_binary(
                self.ctxt, host_name, binary, params_to_update)

            self._compare_obj(result, expected_result)
            mock_update_compute_provider_status.assert_called_once_with(
                self.ctxt, test.MatchType(objects.Service))

    @mock.patch('nova.compute.api.HostAPI._update_compute_provider_status',
                new_callable=mock.NonCallableMock)
    def test_service_update_no_update_provider_status(self, mock_ucps):
        """Tests the scenario that the service is updated but the disabled
        field is not changed, for example the forced_down field is only
        updated. In this case _update_compute_provider_status should not be
        called.
        """
        service = objects.Service(forced_down=True)
        self.assertIn('forced_down', service.obj_what_changed())
        with mock.patch.object(service, 'save') as mock_save:
            retval = self.host_api.service_update(self.ctxt, service)
            self.assertIs(retval, service)
            mock_save.assert_called_once_with()

    @mock.patch('nova.compute.rpcapi.ComputeAPI.set_host_enabled',
                new_callable=mock.NonCallableMock)
    def test_update_compute_provider_status_service_too_old(self, mock_she):
        """Tests the scenario that the service is up but is too old to sync the
        COMPUTE_STATUS_DISABLED trait.
        """
        service = objects.Service(host='fake-host')
        service.version = compute.MIN_COMPUTE_SYNC_COMPUTE_STATUS_DISABLED - 1
        with mock.patch.object(
                self.host_api.servicegroup_api, 'service_is_up',
                return_value=True) as service_is_up:
            self.host_api._update_compute_provider_status(self.ctxt, service)
            service_is_up.assert_called_once_with(service)
        self.assertIn('Compute service on host fake-host is too old to sync '
                      'the COMPUTE_STATUS_DISABLED trait in Placement.',
                      self.stdlog.logger.output)

    @mock.patch('nova.compute.rpcapi.ComputeAPI.set_host_enabled',
                side_effect=messaging.MessagingTimeout)
    def test_update_compute_provider_status_service_rpc_error(self, mock_she):
        """Tests the scenario that the RPC call to the compute service raised
        some exception.
        """
        service = objects.Service(host='fake-host', disabled=True)
        with mock.patch.object(
                self.host_api.servicegroup_api, 'service_is_up',
                return_value=True) as service_is_up:
            self.host_api._update_compute_provider_status(self.ctxt, service)
            service_is_up.assert_called_once_with(service)
        mock_she.assert_called_once_with(self.ctxt, 'fake-host', False)
        log_output = self.stdlog.logger.output
        self.assertIn('An error occurred while updating the '
                      'COMPUTE_STATUS_DISABLED trait on compute node '
                      'resource providers managed by host fake-host.',
                      log_output)
        self.assertIn('MessagingTimeout', log_output)

    @mock.patch.object(objects.InstanceList, 'get_by_host',
                       return_value = ['fake-responses'])
    def test_instance_get_all_by_host(self, mock_get):
        result = self.host_api.instance_get_all_by_host(self.ctxt,
                                                        'fake-host')
        self.assertEqual(['fake-responses'], result)

    @mock.patch(
        'nova.db.main.api.task_log_get_all', return_value='fake-response')
    def test_task_log_get_all(self, mock_task_log_get_all):
        result = self.host_api.task_log_get_all(self.ctxt, 'fake-name',
                                                'fake-begin', 'fake-end',
                                                host='fake-host',
                                                state='fake-state')
        self.assertEqual('fake-response', result)

    @mock.patch.object(objects.CellMappingList, 'get_all',
                       return_value=objects.CellMappingList(objects=[
                           objects.CellMapping(
                               uuid=uuids.cell1_uuid,
                               transport_url='mq://fake1',
                               database_connection='db://fake1'),
                           objects.CellMapping(
                               uuid=uuids.cell2_uuid,
                               transport_url='mq://fake2',
                               database_connection='db://fake2'),
                           objects.CellMapping(
                               uuid=uuids.cell3_uuid,
                               transport_url='mq://fake3',
                               database_connection='db://fake3')]))
    @mock.patch.object(objects.Service, 'get_by_uuid',
                       side_effect=[
                           exception.ServiceNotFound(
                               service_id=uuids.service_uuid),
                           objects.Service(uuid=uuids.service_uuid)])
    def test_service_get_by_id_using_uuid(self, service_get_by_uuid,
                                          cell_mappings_get_all):
        """Tests that we can lookup a service in the HostAPI using a uuid.
        There are two calls to objects.Service.get_by_uuid and the first
        raises ServiceNotFound so that we ensure we keep looping over the
        cells. We'll find the service in the second cell and break the loop
        so that we don't needlessly check in the third cell.
        """

        def _fake_set_target_cell(ctxt, cell_mapping):
            if cell_mapping:
                # These aren't really what would be set for values but let's
                # keep this simple so we can assert something is set when a
                # mapping is provided.
                ctxt.db_connection = cell_mapping.database_connection
                ctxt.mq_connection = cell_mapping.transport_url

        # We have to override the SingleCellSimple fixture.
        self.useFixture(fixtures.MonkeyPatch(
            'nova.context.set_target_cell', _fake_set_target_cell))
        ctxt = context.get_admin_context()
        self.assertIsNone(ctxt.db_connection)
        self.host_api.service_get_by_id(ctxt, uuids.service_uuid)
        # We should have broken the loop over the cells and set the target cell
        # on the context.
        service_get_by_uuid.assert_has_calls(
            [mock.call(ctxt, uuids.service_uuid)] * 2)
        self.assertEqual('db://fake2', ctxt.db_connection)

    @mock.patch('nova.scheduler.client.report.SchedulerReportClient.'
                'aggregate_remove_host')
    @mock.patch('nova.scheduler.client.report.SchedulerReportClient.'
                'aggregate_add_host')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.HostMapping, 'get_by_host')
    def test_service_delete_compute_in_aggregate(
            self, mock_hm, mock_get_cn, mock_add_host, mock_remove_host):
        compute = objects.Service(self.ctxt,
                                  **{'host': 'fake-compute-host',
                                     'binary': 'nova-compute',
                                     'topic': 'compute',
                                     'report_count': 0})
        compute.create()
        # This is needed because of lazy-loading service.compute_node
        cn = objects.ComputeNode(uuid=uuids.cn, host="fake-compute-host",
                                 hypervisor_hostname="fake-compute-host")
        mock_get_cn.return_value = [cn]
        aggregate = self.aggregate_api.create_aggregate(self.ctxt,
                                                   'aggregate',
                                                   None)
        self.aggregate_api.add_host_to_aggregate(self.ctxt,
                                                 aggregate.id,
                                                 'fake-compute-host')
        mock_add_host.assert_called_once_with(
            mock.ANY, aggregate.uuid, host_name='fake-compute-host')
        self.controller.delete(self.req, compute.id)
        result = self.aggregate_api.get_aggregate(self.ctxt,
                                                  aggregate.id).hosts
        self.assertEqual([], result)
        mock_hm.return_value.destroy.assert_called_once_with()
        mock_remove_host.assert_called_once_with(
            mock.ANY, aggregate.uuid, 'fake-compute-host')

    @mock.patch('nova.db.main.api.compute_node_statistics')
    def test_compute_node_statistics(self, mock_cns):
        # Note this should only be called twice
        mock_cns.side_effect = [
            {'stat1': 1, 'stat2': 4.0},
            {'stat1': 5, 'stat2': 1.2},
        ]
        compute.CELLS = [objects.CellMapping(uuid=uuids.cell1),
                         objects.CellMapping(
                                 uuid=objects.CellMapping.CELL0_UUID),
                         objects.CellMapping(uuid=uuids.cell2)]
        stats = self.host_api.compute_node_statistics(self.ctxt)
        self.assertEqual({'stat1': 6, 'stat2': 5.2}, stats)

    @mock.patch.object(objects.CellMappingList, 'get_all',
                       return_value=objects.CellMappingList(objects=[
                           objects.CellMapping(
                               uuid=objects.CellMapping.CELL0_UUID,
                               transport_url='mq://cell0',
                               database_connection='db://cell0'),
                           objects.CellMapping(
                               uuid=uuids.cell1_uuid,
                               transport_url='mq://fake1',
                               database_connection='db://fake1'),
                           objects.CellMapping(
                               uuid=uuids.cell2_uuid,
                               transport_url='mq://fake2',
                               database_connection='db://fake2')]))
    @mock.patch.object(objects.ComputeNode, 'get_by_uuid',
                       side_effect=[exception.ComputeHostNotFound(
                                        host=uuids.cn_uuid),
                                    objects.ComputeNode(uuid=uuids.cn_uuid)])
    def test_compute_node_get_using_uuid(self, compute_get_by_uuid,
                                         cell_mappings_get_all):
        """Tests that we can lookup a compute node in the HostAPI using a uuid.
        """
        self.host_api.compute_node_get(self.ctxt, uuids.cn_uuid)
        # cell0 should have been skipped, and the compute node wasn't found
        # in cell1 so we checked cell2 and found it
        self.assertEqual(2, compute_get_by_uuid.call_count)
        compute_get_by_uuid.assert_has_calls(
            [mock.call(self.ctxt, uuids.cn_uuid)] * 2)

    @mock.patch.object(objects.CellMappingList, 'get_all',
                       return_value=objects.CellMappingList(objects=[
                           objects.CellMapping(
                               uuid=objects.CellMapping.CELL0_UUID,
                               transport_url='mq://cell0',
                               database_connection='db://cell0'),
                           objects.CellMapping(
                               uuid=uuids.cell1_uuid,
                               transport_url='mq://fake1',
                               database_connection='db://fake1'),
                           objects.CellMapping(
                               uuid=uuids.cell2_uuid,
                               transport_url='mq://fake2',
                               database_connection='db://fake2')]))
    @mock.patch.object(objects.ComputeNode, 'get_by_uuid',
                       side_effect=exception.ComputeHostNotFound(
                           host=uuids.cn_uuid))
    def test_compute_node_get_not_found(self, compute_get_by_uuid,
                                        cell_mappings_get_all):
        """Tests that we can lookup a compute node in the HostAPI using a uuid
        and will fail with ComputeHostNotFound if we didn't find it in any
        cell.
        """
        self.assertRaises(exception.ComputeHostNotFound,
                          self.host_api.compute_node_get,
                          self.ctxt, uuids.cn_uuid)
        # cell0 should have been skipped, and the compute node wasn't found
        # in cell1 or cell2.
        self.assertEqual(2, compute_get_by_uuid.call_count)
        compute_get_by_uuid.assert_has_calls(
            [mock.call(self.ctxt, uuids.cn_uuid)] * 2)


class ComputeAggregateAPITestCase(test.TestCase):
    def setUp(self):
        super(ComputeAggregateAPITestCase, self).setUp()
        self.aggregate_api = compute.AggregateAPI()
        self.ctxt = context.get_admin_context()
        # NOTE(jaypipes): We just mock out the HostNapping and Service object
        # lookups in order to bypass the code that does cell lookup stuff,
        # which isn't germane to these tests
        self.useFixture(
            fixtures.MockPatch('nova.objects.HostMapping.get_by_host'))
        self.useFixture(
            fixtures.MockPatch('nova.context.set_target_cell'))
        mock_service_get_by_compute_host = (
            self.useFixture(
                fixtures.MockPatch(
                    'nova.objects.Service.get_by_compute_host')).mock)
        mock_service_get_by_compute_host.return_value = (
            objects.Service(host='fake-host'))

    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch('nova.scheduler.client.report.SchedulerReportClient.'
                'aggregate_add_host')
    @mock.patch.object(compute.LOG, 'warning')
    def test_aggregate_add_host_placement_missing_provider(
            self, mock_log, mock_pc_add_host, mock_get_all_by_host):
        hostname = 'fake-host'
        mock_get_all_by_host.return_value = objects.ComputeNodeList(
            objects=[objects.ComputeNode(
                     host=hostname,
                     hypervisor_hostname=hostname)])
        err = exception.ResourceProviderNotFound(name_or_uuid=hostname)
        mock_pc_add_host.side_effect = err
        aggregate = self.aggregate_api.create_aggregate(
            self.ctxt, 'aggregate', None)
        self.aggregate_api.add_host_to_aggregate(
            self.ctxt, aggregate.id, hostname)
        # Nothing should blow up in Rocky, but we should get a warning
        msg = ("Failed to associate %s with a placement "
               "aggregate: %s. This may be corrected after running "
               "nova-manage placement sync_aggregates.")
        mock_log.assert_called_with(msg, hostname, err)

    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch('nova.scheduler.client.report.SchedulerReportClient.'
                'aggregate_add_host')
    def test_aggregate_add_host_bad_placement(self, mock_pc_add_host,
                                              mock_get_all_by_host):
        hostname = 'fake-host'
        mock_get_all_by_host.return_value = objects.ComputeNodeList(
            objects=[objects.ComputeNode(
                     host=hostname,
                     hypervisor_hostname=hostname)])
        mock_pc_add_host.side_effect = exception.PlacementAPIConnectFailure
        aggregate = self.aggregate_api.create_aggregate(
            self.ctxt, 'aggregate', None)
        agg_uuid = aggregate.uuid
        self.assertRaises(exception.PlacementAPIConnectFailure,
                          self.aggregate_api.add_host_to_aggregate,
                          self.ctxt, aggregate.id, hostname)
        mock_pc_add_host.assert_called_once_with(
            self.ctxt, agg_uuid, host_name=hostname)

    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch('nova.objects.Aggregate.delete_host')
    @mock.patch('nova.scheduler.client.report.SchedulerReportClient.'
                'aggregate_remove_host')
    def test_aggregate_remove_host_bad_placement(
            self, mock_pc_remove_host, mock_agg_obj_delete_host,
            mock_get_all_by_host):
        hostname = 'fake-host'
        mock_get_all_by_host.return_value = objects.ComputeNodeList(
            objects=[objects.ComputeNode(
                host=hostname,
                hypervisor_hostname=hostname)])
        mock_pc_remove_host.side_effect = exception.PlacementAPIConnectFailure
        aggregate = self.aggregate_api.create_aggregate(
            self.ctxt, 'aggregate', None)
        agg_uuid = aggregate.uuid
        self.assertRaises(exception.PlacementAPIConnectFailure,
                          self.aggregate_api.remove_host_from_aggregate,
                          self.ctxt, aggregate.id, hostname)
        mock_pc_remove_host.assert_called_once_with(
            self.ctxt, agg_uuid, hostname)
        # Make sure mock_agg_obj_delete_host wasn't called since placement
        # should be tried first and failed with a server failure.
        mock_agg_obj_delete_host.assert_not_called()

    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch('nova.objects.Aggregate.delete_host')
    @mock.patch('nova.scheduler.client.report.SchedulerReportClient.'
                'aggregate_remove_host')
    @mock.patch.object(compute.LOG, 'warning')
    def test_aggregate_remove_host_placement_missing_provider(
            self, mock_log, mock_pc_remove_host, mock_agg_obj_delete_host,
            mock_get_all_by_host):
        hostname = 'fake-host'
        mock_get_all_by_host.return_value = objects.ComputeNodeList(
            objects=[objects.ComputeNode(
                host=hostname,
                hypervisor_hostname=hostname)])
        err = exception.ResourceProviderNotFound(name_or_uuid=hostname)
        mock_pc_remove_host.side_effect = err
        aggregate = self.aggregate_api.create_aggregate(
            self.ctxt, 'aggregate', None)
        self.aggregate_api.remove_host_from_aggregate(
            self.ctxt, aggregate.id, hostname)
        # Nothing should blow up in Rocky, but we should get a warning
        msg = ("Failed to remove association of %s with a placement "
               "aggregate: %s.")
        mock_log.assert_called_with(msg, hostname, err)
        # In this case Aggregate.delete_host is still called because the
        # ResourceProviderNotFound error is just logged.
        mock_agg_obj_delete_host.assert_called_once_with(hostname)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from oslo_utils.fixture import uuidsentinel as uuids

from nova.api.openstack.compute import baremetal_nodes
from nova.policies import baremetal_nodes as policies
from nova.policies import base as base_policy
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit.policies import base
from nova.tests.unit.virt.ironic import utils as ironic_utils


class BaremetalNodesPolicyTest(base.BasePolicyTest):
    """Test Baremetal Nodes APIs policies with all possible context.

    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super().setUp()

        self.controller = baremetal_nodes.BareMetalNodeController()
        self.req = fakes.HTTPRequest.blank('')

        # stub out openstacksdk
        self.mock_conn = self.useFixture(
            fixtures.MockPatchObject(self.controller, '_ironic_connection'),
        ).mock

        # With legacy rule and scope check disabled by default, system admin,
        # legacy admin, and project admin will be able to get baremetal nodes.
        self.project_admin_authorized_contexts = [
            self.legacy_admin_context,
            self.system_admin_context,
            self.project_admin_context,
        ]

    def test_index_nodes_policy(self):
        rule_name = "os_compute_api:os-baremetal-nodes:list"
        self.mock_conn.nodes.return_value = iter([])

        self.common_policy_auth(self.project_admin_authorized_contexts,
                                rule_name, self.controller.index,
                                self.req)

    def test_show_node_policy(self):
        rule_name = "os_compute_api:os-baremetal-nodes:show"
        properties = {'cpus': 1, 'memory_mb': 512, 'local_gb': 10}
        node = ironic_utils.get_test_node(properties=properties)
        self.mock_conn.get_node.return_value = node
        self.mock_conn.ports.return_value = iter([])

        self.common_policy_auth(self.project_admin_authorized_contexts,
                                rule_name,
                                self.controller.show,
                                self.req, uuids.fake_id)


class BaremetalNodesNoLegacyNoScopePolicyTest(BaremetalNodesPolicyTest):
    """Test Baremetal Nodes APIs policies with no legacy deprecated rules
    and no scope checks which means new defaults only. In that case
    system admin, legacy admin, and project admin will be able to get
    Baremetal nodes Legacy admin will be allowed as policy is just admin if
    no scope checks.

    """

    without_deprecated_rules = True


class BaremetalNodesScopeTypePolicyTest(BaremetalNodesPolicyTest):
    """Test Baremetal Nodes APIs policies with system scope enabled.

    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scopped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(BaremetalNodesScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # With scope checks enable, only project-scoped admins are
        # able to get baremetal nodes.
        self.project_admin_authorized_contexts = [self.legacy_admin_context,
                                                 self.project_admin_context]


class BNScopeTypeNoLegacyPolicyTest(BaremetalNodesScopeTypePolicyTest):
    """Test Baremetal Nodes APIs policies with no legacy deprecated rules
    and scope checks enabled which means scope + new defaults so
    only system admin is able to get baremetal nodes.
    """

    without_deprecated_rules = True
    rules_without_deprecation = {
        policies.BASE_POLICY_NAME % 'list':
            base_policy.ADMIN,
        policies.BASE_POLICY_NAME % 'show':
            base_policy.ADMIN}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import eventlet
from oslo_utils import fixture as utils_fixture
from oslo_utils.fixture import uuidsentinel as uuids
from oslo_utils import timeutils

from nova.compute import api as compute_api
from nova.compute import claims
from nova.compute import instance_actions
from nova.compute import power_state
from nova.compute import task_states
from nova.compute import utils as compute_utils
from nova.compute import vm_states
import nova.conf
from nova import context
from nova.db.main import api as db
from nova import exception
from nova.network import neutron as neutron_api
from nova import objects
from nova import test
from nova.tests import fixtures
from nova.tests.unit.compute import test_compute


CONF = nova.conf.CONF


def _fake_resources():
    resources = {
        'memory_mb': 2048,
        'memory_mb_used': 0,
        'free_ram_mb': 2048,
        'local_gb': 20,
        'local_gb_used': 0,
        'free_disk_gb': 20,
        'vcpus': 2,
        'vcpus_used': 0
    }
    return objects.ComputeNode(**resources)


class ShelveComputeManagerTestCase(test_compute.BaseTestCase):
    @mock.patch.object(objects.BlockDeviceMappingList, 'get_by_instance_uuid')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_terminate_volume_connections')
    @mock.patch.object(nova.virt.fake.SmallFakeDriver, 'power_off')
    @mock.patch.object(nova.virt.fake.SmallFakeDriver, 'snapshot')
    @mock.patch.object(nova.compute.manager.ComputeManager, '_get_power_state')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_notify_about_instance_usage')
    @mock.patch('nova.compute.utils.notify_about_instance_action')
    def _shelve_instance(self, shelved_offload_time, mock_notify,
                         mock_notify_instance_usage, mock_get_power_state,
                         mock_snapshot, mock_power_off, mock_terminate,
                         mock_get_bdms, clean_shutdown=True,
                         guest_power_state=power_state.RUNNING,
                         accel_uuids=None):
        mock_get_power_state.return_value = 123

        CONF.set_override('shelved_offload_time', shelved_offload_time)
        host = 'fake-mini'
        instance = self._create_fake_instance_obj(
            params={'host': host, 'power_state': guest_power_state})
        image_id = 'fake_image_id'
        host = 'fake-mini'
        self.useFixture(utils_fixture.TimeFixture())
        instance.task_state = task_states.SHELVING
        instance.save()

        fake_bdms = None
        if shelved_offload_time == 0:
            fake_bdms = objects.BlockDeviceMappingList()
            mock_get_bdms.return_value = fake_bdms

        tracking = {'last_state': instance.vm_state}

        def check_save(expected_task_state=None):
            self.assertEqual(123, instance.power_state)
            if tracking['last_state'] == vm_states.ACTIVE:
                if CONF.shelved_offload_time == 0:
                    self.assertEqual(task_states.SHELVING_OFFLOADING,
                                     instance.task_state)
                else:
                    self.assertIsNone(instance.task_state)
                self.assertEqual(vm_states.SHELVED, instance.vm_state)
                self.assertEqual([task_states.SHELVING,
                                  task_states.SHELVING_IMAGE_UPLOADING],
                                 expected_task_state)
                self.assertIn('shelved_at', instance.system_metadata)
                self.assertEqual(image_id,
                                 instance.system_metadata['shelved_image_id'])
                self.assertEqual(host,
                                 instance.system_metadata['shelved_host'])
                tracking['last_state'] = instance.vm_state
            elif (tracking['last_state'] == vm_states.SHELVED and
                  CONF.shelved_offload_time == 0):
                self.assertIsNone(instance.task_state)
                self.assertEqual(vm_states.SHELVED_OFFLOADED,
                                 instance.vm_state)
                self.assertEqual([task_states.SHELVING,
                                  task_states.SHELVING_OFFLOADING],
                                 expected_task_state)
                tracking['last_state'] = instance.vm_state
            elif (tracking['last_state'] == vm_states.SHELVED_OFFLOADED and
                  CONF.shelved_offload_time == 0):
                self.assertIsNone(instance.host)
                self.assertIsNone(instance.node)
                self.assertIsNone(expected_task_state)
            else:
                self.fail('Unexpected save!')

        with test.nested(
                mock.patch.object(instance, 'save'),
                mock.patch.object(self.compute.network_api,
                                  'cleanup_instance_network_on_host')) as (
            mock_save, mock_cleanup
        ):
            mock_save.side_effect = check_save
            self.compute.shelve_instance(self.context, instance,
                                         image_id=image_id,
                                         clean_shutdown=clean_shutdown,
                                         accel_uuids=accel_uuids)
            mock_notify.assert_has_calls([
                mock.call(self.context, instance, 'fake-mini',
                          action='shelve', phase='start', bdms=fake_bdms),
                mock.call(self.context, instance, 'fake-mini',
                          action='shelve', phase='end', bdms=fake_bdms)])

        # prepare expect call lists
        mock_notify_instance_usage_call_list = [
            mock.call(self.context, instance, 'shelve.start'),
            mock.call(self.context, instance, 'shelve.end')]
        mock_power_off_call_list = []
        mock_get_power_state_call_list = [mock.call(instance)]

        if clean_shutdown:
            if guest_power_state == power_state.PAUSED:
                mock_power_off_call_list.append(mock.call(instance, 0, 0))
            else:
                mock_power_off_call_list.append(
                    mock.call(instance, CONF.shutdown_timeout,
                              CONF.compute.shutdown_retry_interval))
        else:
            mock_power_off_call_list.append(mock.call(instance, 0, 0))

        if CONF.shelved_offload_time == 0:
            mock_notify_instance_usage_call_list.extend([
                mock.call(self.context, instance, 'shelve_offload.start'),
                mock.call(self.context, instance, 'shelve_offload.end')])
            mock_power_off_call_list.append(mock.call(instance, 0, 0))
            mock_get_power_state_call_list.append(mock.call(instance))

        mock_notify_instance_usage.assert_has_calls(
            mock_notify_instance_usage_call_list)
        mock_power_off.assert_has_calls(mock_power_off_call_list)
        mock_cleanup.assert_not_called()
        mock_snapshot.assert_called_once_with(self.context, instance,
                                              'fake_image_id', mock.ANY)
        mock_get_power_state.assert_has_calls(mock_get_power_state_call_list)

        if CONF.shelved_offload_time == 0:
            self.assertTrue(mock_terminate.called)

    def test_shelve(self):
        self._shelve_instance(-1)

    def test_shelve_forced_shutdown(self):
        self._shelve_instance(-1, clean_shutdown=False)

    def test_shelve_and_offload(self):
        self._shelve_instance(0)

    def test_shelve_paused_instance(self):
        self._shelve_instance(-1, guest_power_state=power_state.PAUSED)

    @mock.patch('nova.accelerator.cyborg._CyborgClient.'
                'delete_arqs_for_instance')
    def test_shelve_and_offload_with_accel_uuids(self, mock_del_arqs):
        self._shelve_instance(0, accel_uuids=[uuids.fake])
        mock_del_arqs.assert_called_once()

    @mock.patch('nova.accelerator.cyborg._CyborgClient.'
                'delete_arqs_for_instance')
    def test_shelve_and_with_accel_uuids(self, mock_del_arqs):
        self._shelve_instance(-1, accel_uuids=[uuids.fake])
        mock_del_arqs.assert_not_called()

    @mock.patch.object(nova.virt.fake.SmallFakeDriver, 'power_off')
    def test_shelve_offload(self, mock_power_off):
        instance = self._shelve_offload()
        mock_power_off.assert_called_once_with(instance,
            CONF.shutdown_timeout, CONF.compute.shutdown_retry_interval)

    @mock.patch.object(nova.virt.fake.SmallFakeDriver, 'power_off')
    def test_shelve_offload_forced_shutdown(self, mock_power_off):
        instance = self._shelve_offload(clean_shutdown=False)
        mock_power_off.assert_called_once_with(instance, 0, 0)

    @mock.patch.object(neutron_api.API, 'unbind_ports')
    @mock.patch.object(compute_utils, 'EventReporter')
    @mock.patch.object(objects.BlockDeviceMappingList, 'get_by_instance_uuid')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_terminate_volume_connections')
    @mock.patch('nova.compute.resource_tracker.ResourceTracker.'
                'delete_allocation_for_shelve_offloaded_instance')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_update_resource_tracker')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_get_power_state', return_value=123)
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_notify_about_instance_usage')
    @mock.patch('nova.compute.utils.notify_about_instance_action')
    def _shelve_offload(self, mock_notify, mock_notify_instance_usage,
                        mock_get_power_state, mock_update_resource_tracker,
                        mock_delete_alloc, mock_terminate, mock_get_bdms,
                        mock_event, mock_unbind_ports, clean_shutdown=True):
        host = 'fake-mini'
        instance = self._create_fake_instance_obj(params={'host': host})
        instance.task_state = task_states.SHELVING
        instance.save()
        self.useFixture(utils_fixture.TimeFixture())
        fake_bdms = objects.BlockDeviceMappingList()
        mock_get_bdms.return_value = fake_bdms

        def stub_instance_save(inst, *args, **kwargs):
            # If the vm_state is changed to SHELVED_OFFLOADED make sure we
            # have already freed up allocations in placement.
            if inst.vm_state == vm_states.SHELVED_OFFLOADED:
                self.assertTrue(mock_delete_alloc.called,
                                'Allocations must be deleted before the '
                                'vm_status can change to shelved_offloaded.')

        self.stub_out('nova.objects.Instance.save', stub_instance_save)
        self.compute.shelve_offload_instance(self.context, instance,
                                             clean_shutdown=clean_shutdown,
                                             accel_uuids=[])
        mock_notify.assert_has_calls([
            mock.call(self.context, instance, 'fake-mini',
                      action='shelve_offload', phase='start',
                      bdms=fake_bdms),
            mock.call(self.context, instance, 'fake-mini',
                      action='shelve_offload', phase='end',
                      bdms=fake_bdms)])

        self.assertEqual(vm_states.SHELVED_OFFLOADED, instance.vm_state)
        self.assertIsNone(instance.task_state)
        self.assertTrue(mock_terminate.called)

        # prepare expect call lists
        mock_notify_instance_usage_call_list = [
            mock.call(self.context, instance, 'shelve_offload.start'),
            mock.call(self.context, instance, 'shelve_offload.end')]

        mock_notify_instance_usage.assert_has_calls(
            mock_notify_instance_usage_call_list)
        # instance.host is replaced with host because
        # original instance.host is clear after
        # ComputeManager.shelve_offload_instance execute
        mock_get_power_state.assert_called_once_with(instance)
        mock_update_resource_tracker.assert_called_once_with(self.context,
                                                             instance)
        mock_delete_alloc.assert_called_once_with(self.context, instance)
        mock_event.assert_called_once_with(self.context,
                                           'compute_shelve_offload_instance',
                                           CONF.host,
                                           instance.uuid,
                                           graceful_exit=False)

        mock_unbind_ports.assert_called_once_with(
            self.context, mock.ANY, detach=False)

        return instance

    @mock.patch('nova.compute.utils.'
                'update_pci_request_with_placement_allocations',
                new=mock.NonCallableMock())
    @mock.patch('nova.objects.BlockDeviceMappingList.get_by_instance_uuid')
    @mock.patch('nova.compute.utils.notify_about_instance_action')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_notify_about_instance_usage')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_prep_block_device', return_value='fake_bdm')
    @mock.patch.object(nova.virt.fake.SmallFakeDriver, 'spawn')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_get_power_state', return_value=123)
    @mock.patch.object(neutron_api.API, 'setup_instance_network_on_host')
    def test_unshelve(self, mock_setup_network,
                      mock_get_power_state, mock_spawn,
                      mock_prep_block_device, mock_notify_instance_usage,
                      mock_notify_instance_action,
                      mock_get_bdms,
                      accel_uuids=None,
                      instance=None):
        mock_bdms = mock.Mock()
        mock_get_bdms.return_value = mock_bdms
        instance = instance or self._create_fake_instance_obj()
        instance.task_state = task_states.UNSHELVING
        instance.save()
        image = {'id': uuids.image_id}
        node = test_compute.NODENAME
        limits = {}
        filter_properties = {'limits': limits}
        host = 'fake-mini'
        cur_time = timeutils.utcnow()
        # Adding shelved_* keys in system metadata to verify
        # whether those are deleted after unshelve call.
        sys_meta = dict(instance.system_metadata)
        sys_meta['shelved_at'] = cur_time.isoformat()
        sys_meta['shelved_image_id'] = image['id']
        sys_meta['shelved_host'] = host
        instance.system_metadata = sys_meta

        self.deleted_image_id = None

        def fake_delete(self2, ctxt, image_id):
            self.deleted_image_id = image_id

        def fake_claim(context, instance, node, allocations, limits):
            instance.host = self.compute.host
            requests = objects.InstancePCIRequests(requests=[])
            return claims.Claim(context, instance, test_compute.NODENAME,
                                self.rt, _fake_resources(),
                                requests)

        tracking = {
            'last_state': instance.task_state,
            'spawned': False,
        }

        def check_save(expected_task_state=None):
            if tracking['last_state'] == task_states.UNSHELVING:
                if tracking['spawned']:
                    self.assertIsNone(instance.task_state)
                else:
                    self.assertEqual(task_states.SPAWNING, instance.task_state)
                    tracking['spawned'] = True
                tracking['last_state'] == instance.task_state
            elif tracking['last_state'] == task_states.SPAWNING:
                self.assertEqual(vm_states.ACTIVE, instance.vm_state)
                tracking['last_state'] == instance.task_state
            else:
                self.fail('Unexpected save!')

        self.useFixture(fixtures.GlanceFixture(self))
        self.stub_out('nova.tests.fixtures.GlanceFixture.delete', fake_delete)

        with mock.patch.object(self.rt, 'instance_claim',
                               side_effect=fake_claim), \
                 mock.patch.object(instance, 'save') as mock_save:
            mock_save.side_effect = check_save
            self.compute.unshelve_instance(
                self.context, instance, image=image,
                filter_properties=filter_properties,
                node=node, request_spec=objects.RequestSpec(),
                accel_uuids=accel_uuids)

        mock_notify_instance_action.assert_has_calls([
            mock.call(self.context, instance, 'fake-mini',
                      action='unshelve', phase='start', bdms=mock_bdms),
            mock.call(self.context, instance, 'fake-mini',
                      action='unshelve', phase='end', bdms=mock_bdms)])

        # prepare expect call lists
        mock_notify_instance_usage_call_list = [
            mock.call(self.context, instance, 'unshelve.start'),
            mock.call(self.context, instance, 'unshelve.end')]

        mock_notify_instance_usage.assert_has_calls(
            mock_notify_instance_usage_call_list)
        mock_prep_block_device.assert_called_once_with(self.context,
            instance, mock.ANY)
        mock_setup_network.assert_called_once_with(
            self.context, instance, self.compute.host, provider_mappings=None)
        mock_spawn.assert_called_once_with(self.context, instance,
                test.MatchType(objects.ImageMeta), injected_files=[],
                admin_password=None, allocations={}, network_info=[],
                block_device_info='fake_bdm', accel_info=mock.ANY)
        self.mock_get_allocations.assert_called_once_with(self.context,
                                                          instance.uuid)
        mock_get_power_state.assert_called_once_with(instance)

        self.assertNotIn('shelved_at', instance.system_metadata)
        self.assertNotIn('shelved_image_id', instance.system_metadata)
        self.assertNotIn('shelved_host', instance.system_metadata)
        self.assertEqual(image['id'], self.deleted_image_id)
        self.assertEqual(instance.host, self.compute.host)

        self.assertEqual(123, instance.power_state)
        self.assertEqual(vm_states.ACTIVE, instance.vm_state)
        self.assertIsNone(instance.task_state)
        self.assertIsNone(instance.key_data)
        self.assertEqual(self.compute.host, instance.host)
        self.assertFalse(instance.auto_disk_config)

    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_get_bound_arq_resources')
    def test_unshelve_with_arqs(self, mock_get_arqs):
        self.test_unshelve(accel_uuids=[uuids.fake])
        mock_get_arqs.assert_called_once()

    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_nil_out_instance_obj_host_and_node')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_terminate_volume_connections')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_build_resources_cleanup')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_get_bound_arq_resources')
    def test_unshelve_with_arqs_failes_cleanup_resources(
            self, mock_get_arqs, mock_cleanup_resources,
            mock_bdms, mock_nil):
        instance = self._create_fake_instance_obj()

        mock_get_arqs.side_effect = exception.AcceleratorRequestOpFailed(
                instance_uuid=instance.uuid, op='get',
                msg='Failure getting accelerator requests.')

        exc = self.assertRaises(exception.AcceleratorRequestOpFailed,
                                self.test_unshelve,
                                accel_uuids=[uuids.fake],
                                instance=instance)
        self.assertIn('Failure getting accelerator requests.', str(exc))
        mock_cleanup_resources.assert_called_once()

    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_nil_out_instance_obj_host_and_node')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_terminate_volume_connections')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_build_resources_cleanup')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_get_bound_arq_resources')
    def test_unshelve_with_arqs_failes_cleanup_resources_timeout(
            self, mock_get_arqs, mock_cleanup_resources,
            mock_bdms, mock_nil):
        instance = self._create_fake_instance_obj()

        mock_get_arqs.side_effect = eventlet.timeout.Timeout()

        self.assertRaises(eventlet.timeout.Timeout,
                          self.test_unshelve,
                          accel_uuids=[uuids.fake],
                          instance=instance)
        mock_cleanup_resources.assert_called_once()

    @mock.patch('nova.objects.BlockDeviceMappingList.get_by_instance_uuid')
    @mock.patch('nova.compute.utils.notify_about_instance_action')
    @mock.patch.object(nova.compute.resource_tracker.ResourceTracker,
                       'instance_claim')
    @mock.patch.object(neutron_api.API, 'setup_instance_network_on_host')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_get_power_state', return_value=123)
    @mock.patch.object(nova.virt.fake.SmallFakeDriver, 'spawn')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_prep_block_device', return_value='fake_bdm')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_notify_about_instance_usage')
    @mock.patch('nova.utils.get_image_from_system_metadata')
    def test_unshelve_volume_backed(self, mock_image_meta,
                                    mock_notify_instance_usage,
                                    mock_prep_block_device, mock_spawn,
                                    mock_get_power_state,
                                    mock_setup_network, mock_instance_claim,
                                    mock_notify_instance_action,
                                    mock_get_bdms):
        mock_bdms = mock.Mock()
        mock_get_bdms.return_value = mock_bdms
        instance = self._create_fake_instance_obj()
        node = test_compute.NODENAME
        limits = {}
        filter_properties = {'limits': limits}
        instance.task_state = task_states.UNSHELVING
        instance.save()
        image_meta = {'properties': {'base_image_ref': uuids.image_id}}
        mock_image_meta.return_value = image_meta

        tracking = {'last_state': instance.task_state}

        def fake_claim(context, instance, node, allocations, limits):
            instance.host = self.compute.host
            requests = objects.InstancePCIRequests(requests=[])
            return claims.Claim(context, instance, test_compute.NODENAME,
                                self.rt, _fake_resources(),
                                requests)
        mock_instance_claim.side_effect = fake_claim

        def check_save(expected_task_state=None):
            if tracking['last_state'] == task_states.UNSHELVING:
                self.assertEqual(task_states.SPAWNING, instance.task_state)
                tracking['last_state'] = instance.task_state
            elif tracking['last_state'] == task_states.SPAWNING:
                self.assertEqual(123, instance.power_state)
                self.assertEqual(vm_states.ACTIVE, instance.vm_state)
                self.assertIsNone(instance.task_state)
                self.assertIsNone(instance.key_data)
                self.assertFalse(instance.auto_disk_config)
                self.assertIsNone(instance.task_state)
                tracking['last_state'] = instance.task_state
            else:
                self.fail('Unexpected save!')

        with mock.patch.object(instance, 'save') as mock_save:
            mock_save.side_effect = check_save
            self.compute.unshelve_instance(self.context, instance, image=None,
                    filter_properties=filter_properties, node=node,
                    request_spec=objects.RequestSpec(), accel_uuids=[])

        mock_notify_instance_action.assert_has_calls([
            mock.call(self.context, instance, 'fake-mini',
                      action='unshelve', phase='start', bdms=mock_bdms),
            mock.call(self.context, instance, 'fake-mini',
                      action='unshelve', phase='end', bdms=mock_bdms)])

        # prepare expect call lists
        mock_notify_instance_usage_call_list = [
            mock.call(self.context, instance, 'unshelve.start'),
            mock.call(self.context, instance, 'unshelve.end')]

        mock_notify_instance_usage.assert_has_calls(
            mock_notify_instance_usage_call_list)
        mock_prep_block_device.assert_called_once_with(self.context, instance,
                mock.ANY)
        mock_setup_network.assert_called_once_with(
            self.context, instance, self.compute.host, provider_mappings=None)
        mock_instance_claim.assert_called_once_with(self.context, instance,
                                                    test_compute.NODENAME,
                                                    {}, limits)
        mock_spawn.assert_called_once_with(self.context, instance,
                test.MatchType(objects.ImageMeta),
                injected_files=[], admin_password=None, accel_info=[],
                allocations={}, network_info=[], block_device_info='fake_bdm')
        self.mock_get_allocations.assert_called_once_with(self.context,
                                                          instance.uuid)
        mock_get_power_state.assert_called_once_with(instance)

    @mock.patch('nova.objects.BlockDeviceMappingList.get_by_instance_uuid')
    @mock.patch('nova.compute.utils.notify_about_instance_action')
    @mock.patch.object(nova.compute.resource_tracker.ResourceTracker,
                       'instance_claim')
    @mock.patch.object(neutron_api.API, 'setup_instance_network_on_host')
    @mock.patch.object(nova.virt.fake.SmallFakeDriver, 'spawn',
                       side_effect=test.TestingException('oops!'))
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_prep_block_device', return_value='fake_bdm')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_notify_about_instance_usage')
    @mock.patch('nova.utils.get_image_from_system_metadata')
    @mock.patch.object(nova.compute.manager.ComputeManager,
                       '_terminate_volume_connections')
    def test_unshelve_spawn_fails_cleanup_volume_connections(
            self, mock_terminate_volume_connections, mock_image_meta,
            mock_notify_instance_usage, mock_prep_block_device, mock_spawn,
            mock_setup_network, mock_instance_claim,
            mock_notify_instance_action, mock_get_bdms):
        """Tests error handling when a instance fails to unshelve and makes
        sure that volume connections are cleaned up from the host
        and that the host/node values are unset on the instance.
        """
        mock_bdms = mock.Mock()
        mock_get_bdms.return_value = mock_bdms
        instance = self._create_fake_instance_obj()
        node = test_compute.NODENAME
        limits = {}
        filter_properties = {'limits': limits}
        instance.task_state = task_states.UNSHELVING
        instance.save()
        image_meta = {'properties': {'base_image_ref': uuids.image_id}}
        mock_image_meta.return_value = image_meta

        tracking = {'last_state': instance.task_state}

        def fake_claim(context, instance, node, allocations, limits):
            instance.host = self.compute.host
            instance.node = node
            requests = objects.InstancePCIRequests(requests=[])
            return claims.Claim(context, instance, node,
                                self.rt, _fake_resources(),
                                requests, limits=limits)
        mock_instance_claim.side_effect = fake_claim

        def check_save(expected_task_state=None):
            if tracking['last_state'] == task_states.UNSHELVING:
                # This is before we've failed.
                self.assertEqual(task_states.SPAWNING, instance.task_state)
                tracking['last_state'] = instance.task_state
            elif tracking['last_state'] == task_states.SPAWNING:
                # This is after we've failed.
                self.assertIsNone(instance.host)
                self.assertIsNone(instance.node)
                self.assertIsNone(instance.task_state)
                tracking['last_state'] = instance.task_state
            else:
                self.fail('Unexpected save!')

        with mock.patch.object(instance, 'save') as mock_save:
            mock_save.side_effect = check_save
            self.assertRaises(test.TestingException,
                              self.compute.unshelve_instance,
                              self.context, instance, image=None,
                              filter_properties=filter_properties, node=node,
                              request_spec=objects.RequestSpec(),
                              accel_uuids=[])

        mock_notify_instance_action.assert_called_once_with(
            self.context, instance, 'fake-mini', action='unshelve',
            phase='start', bdms=mock_bdms)
        mock_notify_instance_usage.assert_called_once_with(
            self.context, instance, 'unshelve.start')
        mock_prep_block_device.assert_called_once_with(
            self.context, instance, mock_bdms)
        mock_setup_network.assert_called_once_with(
            self.context, instance, self.compute.host, provider_mappings=None)
        mock_instance_claim.assert_called_once_with(self.context, instance,
                                                    test_compute.NODENAME,
                                                    {}, limits)
        mock_spawn.assert_called_once_with(
            self.context, instance, test.MatchType(objects.ImageMeta),
            injected_files=[], admin_password=None, accel_info=[],
            allocations={}, network_info=[], block_device_info='fake_bdm')
        mock_terminate_volume_connections.assert_called_once_with(
            self.context, instance, mock_bdms)

    @mock.patch('nova.network.neutron.API.setup_instance_network_on_host')
    @mock.patch('nova.compute.utils.'
                'update_pci_request_with_placement_allocations')
    def test_unshelve_with_resource_request(
            self, mock_update_pci, mock_setup_network):
        requested_res = [objects.RequestGroup(
            requester_id=uuids.port_1,
            provider_uuids=[uuids.rp1])]
        request_spec = objects.RequestSpec(requested_resources=requested_res)
        instance = self._create_fake_instance_obj()

        self.compute.unshelve_instance(
            self.context, instance, image=None,
            filter_properties={}, node='fakenode1', request_spec=request_spec,
            accel_uuids=[])

        mock_update_pci.assert_called_once_with(
            self.context, self.compute.reportclient, [],
            {uuids.port_1: [uuids.rp1]})
        mock_setup_network.assert_called_once_with(
            self.context, instance, self.compute.host,
            provider_mappings={uuids.port_1: [uuids.rp1]})

    @mock.patch('nova.network.neutron.API.setup_instance_network_on_host',
                new=mock.NonCallableMock())
    @mock.patch('nova.compute.utils.'
                'update_pci_request_with_placement_allocations')
    def test_unshelve_with_resource_request_update_raises(
            self, mock_update_pci):
        requested_res = [objects.RequestGroup(
            requester_id=uuids.port_1,
            provider_uuids=[uuids.rp1])]
        request_spec = objects.RequestSpec(requested_resources=requested_res)
        instance = self._create_fake_instance_obj()
        mock_update_pci.side_effect = (
            exception.UnexpectedResourceProviderNameForPCIRequest(
                provider=uuids.rp1,
                requester=uuids.port1,
                provider_name='unexpected'))

        self.assertRaises(
            exception.UnexpectedResourceProviderNameForPCIRequest,
            self.compute.unshelve_instance, self.context, instance, image=None,
            filter_properties={}, node='fake-node', request_spec=request_spec,
            accel_uuids=[])

        mock_update_pci.assert_called_once_with(
            self.context, self.compute.reportclient, [],
            {uuids.port_1: [uuids.rp1]})

    def test_unshelve_spawn_fails_cleanup_instance_image_ref(self):
        """Tests error handling when a instance fails to unshelve and makes
        sure to revert instance.image_ref to the initial value instead of
        keeping the one of shelved image
        https://bugs.launchpad.net/nova/+bug/1934094
        """
        instance = self._create_fake_instance_obj()
        initial_image_ref = instance.image_ref

        fake_spec = objects.RequestSpec()
        shelved_image = {'id': uuids.image_id}

        with mock.patch.object(self.compute.driver, 'spawn',
                side_effect=test.TestingException('Spawn Failed')):
            self.assertRaises(test.TestingException,
                    self.compute.unshelve_instance, self.context, instance,
                    image=shelved_image, filter_properties={},
                    node='fakenode1', request_spec=fake_spec, accel_uuids=[])
        self.assertEqual(instance.image_ref, initial_image_ref)

    @mock.patch.object(objects.InstanceList, 'get_by_filters')
    def test_shelved_poll_none_offloaded(self, mock_get_by_filters):
        # Test instances are not offloaded when shelved_offload_time is -1
        self.flags(shelved_offload_time=-1)
        self.compute._poll_shelved_instances(self.context)
        self.assertEqual(0, mock_get_by_filters.call_count)

    @mock.patch('oslo_utils.timeutils.is_older_than')
    def test_shelved_poll_none_exist(self, mock_older):
        self.flags(shelved_offload_time=1)
        mock_older.return_value = False

        with mock.patch.object(self.compute, 'shelve_offload_instance') as soi:
            self.compute._poll_shelved_instances(self.context)
            self.assertFalse(soi.called)

    @mock.patch('oslo_utils.timeutils.is_older_than')
    def test_shelved_poll_not_timedout(self, mock_older):
        mock_older.return_value = False
        self.flags(shelved_offload_time=1)
        shelved_time = timeutils.utcnow()
        time_fixture = self.useFixture(utils_fixture.TimeFixture(shelved_time))
        time_fixture.advance_time_seconds(CONF.shelved_offload_time - 1)
        instance = self._create_fake_instance_obj()
        instance.vm_state = vm_states.SHELVED
        instance.task_state = None
        instance.host = self.compute.host
        sys_meta = instance.system_metadata
        sys_meta['shelved_at'] = shelved_time.isoformat()
        instance.save()

        with mock.patch.object(self.compute, 'shelve_offload_instance') as soi:
            self.compute._poll_shelved_instances(self.context)
            self.assertFalse(soi.called)
            self.assertTrue(mock_older.called)

    def test_shelved_poll_timedout(self):
        self.flags(shelved_offload_time=1)
        shelved_time = timeutils.utcnow()
        time_fixture = self.useFixture(utils_fixture.TimeFixture(shelved_time))
        time_fixture.advance_time_seconds(CONF.shelved_offload_time + 1)
        instance = self._create_fake_instance_obj()
        instance.vm_state = vm_states.SHELVED
        instance.task_state = None
        instance.host = self.compute.host
        sys_meta = instance.system_metadata
        sys_meta['shelved_at'] = shelved_time.isoformat()
        instance.save()

        data = []

        def fake_soi(context, instance, accel_uuids, **kwargs):
            data.append(instance.uuid)

        with mock.patch.object(self.compute, 'shelve_offload_instance') as soi:
            soi.side_effect = fake_soi
            self.compute._poll_shelved_instances(self.context)
            self.assertTrue(soi.called)
            self.assertEqual(instance.uuid, data[0])

    @mock.patch('oslo_utils.timeutils.is_older_than')
    @mock.patch('oslo_utils.timeutils.parse_strtime')
    def test_shelved_poll_filters_task_state(self, mock_parse, mock_older):
        self.flags(shelved_offload_time=1)
        mock_older.return_value = True
        instance1 = self._create_fake_instance_obj()
        instance1.task_state = task_states.SPAWNING
        instance1.vm_state = vm_states.SHELVED
        instance1.host = self.compute.host
        instance1.system_metadata = {'shelved_at': ''}
        instance1.save()
        instance2 = self._create_fake_instance_obj()
        instance2.task_state = None
        instance2.vm_state = vm_states.SHELVED
        instance2.host = self.compute.host
        instance2.system_metadata = {'shelved_at': ''}
        instance2.save()

        data = []

        def fake_soi(context, instance, accel_uuids, **kwargs):
            data.append(instance.uuid)

        with mock.patch.object(self.compute, 'shelve_offload_instance') as soi:
            soi.side_effect = fake_soi
            self.compute._poll_shelved_instances(self.context)
            self.assertTrue(soi.called)
            self.assertEqual([instance2.uuid], data)

    @mock.patch('nova.accelerator.cyborg._CyborgClient.'
                'get_arq_uuids_for_instance')
    @mock.patch('oslo_utils.timeutils.is_older_than')
    @mock.patch('oslo_utils.timeutils.parse_strtime')
    def test_shelved_poll_with_accel_uuids(self, mock_parse, mock_older,
                                           mock_get_arq_uuids):
        self.flags(shelved_offload_time=1)
        mock_older.return_value = True
        instance = self._create_fake_instance_obj()
        instance.task_state = None
        instance.vm_state = vm_states.SHELVED
        instance.host = self.compute.host
        instance.system_metadata = {'shelved_at': ''}
        instance.flavor.extra_specs['accel:device_profile'] = 'dp_test'
        mock_get_arq_uuids.return_value = [uuids.fake]
        instance.save()

        data = []

        def fake_soi(context, instance, **kwargs):
            data.append(instance.uuid)

        with mock.patch.object(self.compute, 'shelve_offload_instance') as soi:
            soi.side_effect = fake_soi
            self.compute._poll_shelved_instances(self.context)
            self.assertTrue(soi.called)
            self.assertEqual([instance.uuid], data)

    @mock.patch('oslo_utils.timeutils.is_older_than')
    @mock.patch('oslo_utils.timeutils.parse_strtime')
    def test_shelved_poll_checks_task_state_on_save(self, mock_parse,
                                                    mock_older):
        self.flags(shelved_offload_time=1)
        mock_older.return_value = True
        instance = self._create_fake_instance_obj()
        instance.task_state = None
        instance.vm_state = vm_states.SHELVED
        instance.host = self.compute.host
        instance.system_metadata = {'shelved_at': ''}
        instance.save()

        def fake_parse_hook(timestring):
            instance.task_state = task_states.SPAWNING
            instance.save()

        mock_parse.side_effect = fake_parse_hook

        with mock.patch.object(self.compute, 'shelve_offload_instance') as soi:
            self.compute._poll_shelved_instances(self.context)
            self.assertFalse(soi.called)


class ShelveComputeAPITestCase(test_compute.BaseTestCase):
    def _get_vm_states(self, exclude_states=None):
        vm_state = set([vm_states.ACTIVE, vm_states.BUILDING, vm_states.PAUSED,
                    vm_states.SUSPENDED, vm_states.RESCUED, vm_states.STOPPED,
                    vm_states.RESIZED, vm_states.SOFT_DELETED,
                    vm_states.DELETED, vm_states.ERROR, vm_states.SHELVED,
                    vm_states.SHELVED_OFFLOADED])
        if not exclude_states:
            exclude_states = set()
        return vm_state - exclude_states

    @mock.patch('nova.scheduler.client.report.SchedulerReportClient.'
                'aggregate_add_host')
    @mock.patch('nova.availability_zones.get_availability_zones')
    def _create_host_inside_az(
            self,
            ctxt,
            host,
            az,
            mock_az,
            mock_aggregate,
            ):

        self.api = compute_api.AggregateAPI()
        mock_az.return_value = [az]

        cells = objects.CellMappingList.get_all(ctxt)
        cell = cells[0]
        with context.target_cell(ctxt, cell) as cctxt:
            s = objects.Service(context=cctxt,
                                host=host,
                                binary='nova-compute',
                                topic='compute',
                                report_count=0)
            s.create()

        hm = objects.HostMapping(context=ctxt,
                                 cell_mapping=cell,
                                 host=host)
        hm.create()

        self._init_aggregate_with_host(None, 'fake_aggregate1',
                                       az, host)

    def _create_request_spec_for_initial_az(self, az):
        fake_spec = objects.RequestSpec()
        fake_spec.availability_zone = az
        return fake_spec

    def _assert_unshelving_and_request_spec_az_and_host(
            self,
            context,
            instance,
            fake_spec,
            fake_zone,
            fake_host,
            mock_get_by_instance_uuid,
            mock_unshelve
    ):
        mock_get_by_instance_uuid.assert_called_once_with(context,
                                                     instance.uuid)

        mock_unshelve.assert_called_once_with(context, instance, fake_spec)

        self.assertEqual(instance.task_state, task_states.UNSHELVING)
        self.assertEqual(fake_spec.availability_zone, fake_zone)
        if fake_host:
            self.assertEqual(fake_spec.requested_destination.host, fake_host)

    def _test_shelve(self, vm_state=vm_states.ACTIVE, boot_from_volume=False,
                     clean_shutdown=True):

        params = dict(task_state=None, vm_state=vm_state, display_name='vm01')
        fake_instance = self._create_fake_instance_obj(params=params)
        instance = fake_instance

        self.assertIsNone(instance['task_state'])

        with test.nested(
            mock.patch.object(compute_utils, 'is_volume_backed_instance',
                              return_value=boot_from_volume),
            mock.patch.object(compute_utils, 'create_image',
                              return_value=dict(id='fake-image-id')),
            mock.patch.object(instance, 'save'),
            mock.patch.object(self.compute_api, '_record_action_start'),
            mock.patch.object(self.compute_api.compute_rpcapi,
                              'shelve_instance'),
            mock.patch.object(self.compute_api.compute_rpcapi,
                              'shelve_offload_instance')
        ) as (
            volume_backed_inst, create_image, instance_save,
            record_action_start, rpcapi_shelve_instance,
            rpcapi_shelve_offload_instance
        ):

            self.compute_api.shelve(self.context, instance,
                                    clean_shutdown=clean_shutdown)

            self.assertEqual(instance.task_state, task_states.SHELVING)
            # assert our mock calls
            volume_backed_inst.assert_called_once_with(
                self.context, instance)
            instance_save.assert_called_once_with(expected_task_state=[None])
            record_action_start.assert_called_once_with(
                self.context, instance, instance_actions.SHELVE)
            if boot_from_volume:
                rpcapi_shelve_offload_instance.assert_called_once_with(
                    self.context, instance=instance,
                    clean_shutdown=clean_shutdown, accel_uuids=[])
            else:
                rpcapi_shelve_instance.assert_called_once_with(
                    self.context, instance=instance, image_id='fake-image-id',
                    clean_shutdown=clean_shutdown, accel_uuids=[])

            db.instance_destroy(self.context, instance['uuid'])

    def test_shelve(self):
        self._test_shelve()

    def test_shelves_stopped(self):
        self._test_shelve(vm_state=vm_states.STOPPED)

    def test_shelves_paused(self):
        self._test_shelve(vm_state=vm_states.PAUSED)

    def test_shelves_suspended(self):
        self._test_shelve(vm_state=vm_states.SUSPENDED)

    def test_shelves_boot_from_volume(self):
        self._test_shelve(boot_from_volume=True)

    def test_shelve_forced_shutdown(self):
        self._test_shelve(clean_shutdown=False)

    def test_shelve_boot_from_volume_forced_shutdown(self):
        self._test_shelve(boot_from_volume=True,
                          clean_shutdown=False)

    def _test_shelve_invalid_state(self, vm_state):
        params = dict(vm_state=vm_state)
        fake_instance = self._create_fake_instance_obj(params=params)
        self.assertRaises(exception.InstanceInvalidState,
                          self.compute_api.shelve,
                          self.context, fake_instance)

    def test_shelve_fails_invalid_states(self):
        invalid_vm_states = self._get_vm_states(set([vm_states.ACTIVE,
                                                     vm_states.STOPPED,
                                                     vm_states.PAUSED,
                                                     vm_states.SUSPENDED]))
        for state in invalid_vm_states:
            self._test_shelve_invalid_state(state)

    def _test_shelve_offload(self, clean_shutdown=True):
        params = dict(task_state=None, vm_state=vm_states.SHELVED)
        fake_instance = self._create_fake_instance_obj(params=params)
        with test.nested(
            mock.patch.object(fake_instance, 'save'),
            mock.patch.object(self.compute_api.compute_rpcapi,
                              'shelve_offload_instance'),
            mock.patch('nova.compute.api.API._record_action_start')
        ) as (
            instance_save, rpcapi_shelve_offload_instance, record
        ):
            self.compute_api.shelve_offload(self.context, fake_instance,
                                            clean_shutdown=clean_shutdown)
            # assert field values set on the instance object
            self.assertEqual(task_states.SHELVING_OFFLOADING,
                             fake_instance.task_state)
            instance_save.assert_called_once_with(expected_task_state=[None])
            rpcapi_shelve_offload_instance.assert_called_once_with(
                    self.context, instance=fake_instance,
                    clean_shutdown=clean_shutdown, accel_uuids=[])
            record.assert_called_once_with(self.context, fake_instance,
                                           instance_actions.SHELVE_OFFLOAD)

    def test_shelve_offload(self):
        self._test_shelve_offload()

    def test_shelve_offload_forced_shutdown(self):
        self._test_shelve_offload(clean_shutdown=False)

    def _test_shelve_offload_invalid_state(self, vm_state):
        params = dict(vm_state=vm_state)
        fake_instance = self._create_fake_instance_obj(params=params)
        self.assertRaises(exception.InstanceInvalidState,
                          self.compute_api.shelve_offload,
                          self.context, fake_instance)

    def test_shelve_offload_fails_invalid_states(self):
        invalid_vm_states = self._get_vm_states(set([vm_states.SHELVED]))
        for state in invalid_vm_states:
            self._test_shelve_offload_invalid_state(state)

    def _get_specify_state_instance(self, vm_state):
        # Ensure instance can be unshelved.
        instance = self._create_fake_instance_obj()

        self.assertIsNone(instance['task_state'])

        self.compute_api.shelve(self.context, instance)

        instance.task_state = None
        instance.vm_state = vm_state
        instance.save()

        return instance

    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve(self, get_by_instance_uuid, fake_save):
        # Ensure instance can be unshelved.
        instance = self._get_specify_state_instance(vm_states.SHELVED)

        fake_spec = objects.RequestSpec()
        fake_spec.availability_zone = None
        get_by_instance_uuid.return_value = fake_spec
        with mock.patch.object(self.compute_api.compute_task_api,
                               'unshelve_instance') as unshelve:
            self.compute_api.unshelve(self.context, instance)
            get_by_instance_uuid.assert_called_once_with(self.context,
                                                         instance.uuid)
            unshelve.assert_called_once_with(self.context, instance, fake_spec)

        self.assertEqual(instance.task_state, task_states.UNSHELVING)

        db.instance_destroy(self.context, instance['uuid'])

    @mock.patch('nova.availability_zones.get_availability_zones',
                return_value=['az1', 'az2'])
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_specified_az_ushelve_invalid_request(self,
                                                  get_by_instance_uuid,
                                                  mock_save,
                                                  mock_availability_zones):
        # Ensure instance can be unshelved.
        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        new_az = "fake-new-az"
        fake_spec = objects.RequestSpec()
        fake_spec.availability_zone = "fake-old-az"
        get_by_instance_uuid.return_value = fake_spec

        exc = self.assertRaises(exception.InvalidRequest,
                                self.compute_api.unshelve,
                                self.context, instance, new_az=new_az)
        self.assertEqual("The requested availability zone is not available",
                         exc.format_message())

    @mock.patch('nova.availability_zones.get_availability_zones',
                return_value=['az1', 'az2'])
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_specified_az_unshelve_invalid_state(self, get_by_instance_uuid,
                                                 mock_save,
                                                 mock_availability_zones):
        # Ensure instance can be unshelved.
        instance = self._get_specify_state_instance(vm_states.SHELVED)

        new_az = "az1"
        fake_spec = objects.RequestSpec()
        fake_spec.availability_zone = "fake-old-az"
        get_by_instance_uuid.return_value = fake_spec

        self.assertRaises(exception.UnshelveInstanceInvalidState,
                          self.compute_api.unshelve,
                          self.context, instance, new_az=new_az)

    @mock.patch(
        'nova.objects.service.get_minimum_version_all_cells',
        new=mock.Mock(return_value=58),
    )
    @mock.patch(
        'nova.network.neutron.API.instance_has_extended_resource_request',
        new=mock.Mock(return_value=True),
    )
    def test_unshelve_offloaded_with_extended_resource_request_old_compute(
        self
    ):
        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        self.assertRaises(
            exception.ExtendedResourceRequestOldCompute,
            self.compute_api.unshelve, self.context, instance, new_az="az1"
        )

    @mock.patch('nova.objects.BlockDeviceMappingList.get_by_instance_uuid',
                new_callable=mock.NonCallableMock)
    @mock.patch('nova.availability_zones.get_availability_zones')
    def test_validate_unshelve_az_cross_az_attach_true(
            self, mock_get_azs, mock_get_bdms):
        """Tests a case where the new AZ to unshelve does not match the volume
        attached to the server but cross_az_attach=True so it's not an error.
        """
        # Ensure instance can be unshelved.
        instance = self._create_fake_instance_obj(
            params=dict(vm_state=vm_states.SHELVED_OFFLOADED))

        new_az = "west_az"
        mock_get_azs.return_value = ["west_az", "east_az"]
        self.flags(cross_az_attach=True, group='cinder')
        self.compute_api._validate_unshelve_az(self.context, instance, new_az)
        mock_get_azs.assert_called_once_with(
            self.context, self.compute_api.host_api, get_only_available=True)

    @mock.patch('nova.volume.cinder.API.get')
    @mock.patch('nova.objects.BlockDeviceMappingList.get_by_instance_uuid')
    @mock.patch('nova.availability_zones.get_availability_zones')
    def test_validate_unshelve_az_cross_az_attach_false(
            self, mock_get_azs, mock_get_bdms, mock_get):
        """Tests a case where the new AZ to unshelve does not match the volume
        attached to the server and cross_az_attach=False so it's an error.
        """
        # Ensure instance can be unshelved.
        instance = self._create_fake_instance_obj(
            params=dict(vm_state=vm_states.SHELVED_OFFLOADED))

        new_az = "west_az"
        mock_get_azs.return_value = ["west_az", "east_az"]

        bdms = [objects.BlockDeviceMapping(destination_type='volume',
                                           volume_id=uuids.volume_id)]
        mock_get_bdms.return_value = bdms
        volume = {'id': uuids.volume_id, 'availability_zone': 'east_az'}
        mock_get.return_value = volume

        self.flags(cross_az_attach=False, group='cinder')
        self.assertRaises(exception.MismatchVolumeAZException,
                          self.compute_api._validate_unshelve_az,
                          self.context, instance, new_az)
        mock_get_azs.assert_called_once_with(
            self.context, self.compute_api.host_api, get_only_available=True)
        mock_get_bdms.assert_called_once_with(self.context, instance.uuid)
        mock_get.assert_called_once_with(self.context, uuids.volume_id)

# Next tests attempt to check the following behavior
# +----------+---------------------------+-------+----------------------------+
# | Boot     | Unshelve after offload AZ | Host  | Result                     |
# +==========+===========================+=======+============================+
# |  No AZ   | No AZ or AZ=null          | No    | Free scheduling,           |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
# |  No AZ   | No AZ or AZ=null          | Host1 | Schedule to host1,         |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
# |  No AZ   | AZ="AZ1"                  | No    | Schedule to AZ1,           |
# |          |                           |       | reqspec.AZ="AZ1"           |
# +----------+---------------------------+-------+----------------------------+
# |  No AZ   | AZ="AZ1"                  | Host1 | Verify that host1 in AZ1,  |
# |          |                           |       | or (1). Schedule to        |
# |          |                           |       | host1, reqspec.AZ="AZ1"    |
# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | No AZ                     | No    | Schedule to AZ1,           |
# |          |                           |       | reqspec.AZ="AZ1"           |
# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ=null                   | No    | Free scheduling,           |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | No AZ                     | Host1 | If host1 is in AZ1,        |
# |          |                           |       | then schedule to host1,    |
# |          |                           |       | reqspec.AZ="AZ1", otherwise|
# |          |                           |       | reject the request (1)     |
# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ=null                   | Host1 | Schedule to host1,         |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ="AZ2"                  | No    | Schedule to AZ2,           |
# |          |                           |       | reqspec.AZ="AZ2"           |
# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ="AZ2"                  | Host1 | If host1 in AZ2 then       |
# |          |                           |       | schedule to host1,         |
# |          |                           |       | reqspec.AZ="AZ2",          |
# |          |                           |       | otherwise reject (1)       |
# +----------+---------------------------+-------+----------------------------+
#
# (1) Check at the api and return an error.
#
#
# +----------+---------------------------+-------+----------------------------+
# |  No AZ   | No AZ or AZ=null          | No    | Free scheduling,           |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_without_az(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(None)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(context, instance)

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            None,
            None,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

# +----------+---------------------------+-------+----------------------------+
# |  No AZ   | No AZ or AZ=null          | Host1 | Schedule to host1,         |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_without_az_to_host(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(None)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(context, instance, host=fake_host)

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            None,
            fake_host,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

# +----------+---------------------------+-------+----------------------------+
# |  No AZ   | AZ="AZ1"                  | No    | Schedule to AZ1,           |
# |          |                           |       | reqspec.AZ="AZ1"           |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_without_az_to_newaz(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(None)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(context, instance, new_az=fake_zone)

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            fake_zone,
            None,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

# +----------+---------------------------+-------+----------------------------+
# |  No AZ   | AZ="AZ1"                  | Host1 | Verify that host1 in AZ1,  |
# |          |                           |       | or (1). Schedule to        |
# |          |                           |       | host1, reqspec.AZ="AZ1"    |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_without_az_to_newaz_and_host(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(None)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(
            context, instance, new_az=fake_zone, host=fake_host
        )

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            fake_zone,
            fake_host,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_without_az_to_newaz_and_host_invalid(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(None)
        mock_get_by_instance_uuid.return_value = fake_spec

        exc = self.assertRaises(
            nova.exception.UnshelveHostNotInAZ,
            self.compute_api.unshelve,
            context,
            instance,
            new_az='avail_zone1',
            host='fake_mini'
        )

        self.assertIn(
            exc.message,
            'Host "fake_mini" is not in the availability zone "avail_zone1".'
        )

# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | No AZ                     | No    | Schedule to AZ1,           |
# |          |                           |       | reqspec.AZ="AZ1"           |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(fake_zone)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(context, instance)

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            fake_zone,
            None,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ=null                   | No    | Free scheduling,           |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az_to_unpin_az(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(fake_zone)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(context, instance, new_az=None)

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            None,
            None,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | No AZ                     | Host1 | If host1 is in AZ1,        |
# |          |                           |       | then schedule to host1,    |
# |          |                           |       | reqspec.AZ="AZ1", otherwise|
# |          |                           |       | reject the request (1)     |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az_to_host_in_az(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(fake_zone)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(context, instance, host=fake_host)

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            fake_zone,
            fake_host,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az_to_invalid_host(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(fake_zone)
        mock_get_by_instance_uuid.return_value = fake_spec

        exc = self.assertRaises(
            nova.exception.UnshelveHostNotInAZ,
            self.compute_api.unshelve,
            context,
            instance,
            host='fake_mini'
        )

        self.assertIn(
            exc.message,
            'Host "fake_mini" is not in the availability zone "avail_zone1".'
        )

# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ=null                   | Host1 | Schedule to host1,         |
# |          |                           |       | reqspec.AZ=None            |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az_to_host_unpin_az(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az(fake_zone)
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(
            context, instance, new_az=None, host=fake_host
        )

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            None,
            fake_host,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ="AZ2"                  | No    | Schedule to AZ2,           |
# |          |                           |       | reqspec.AZ="AZ2"           |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az_to_newaz(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az('az1')
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(
            context, instance, new_az=fake_zone
        )

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            fake_zone,
            None,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

# +----------+---------------------------+-------+----------------------------+
# |  AZ1     | AZ="AZ2"                  | Host1 | If host1 in AZ2 then       |
# |          |                           |       | schedule to host1,         |
# |          |                           |       | reqspec.AZ="AZ2",          |
# |          |                           |       | otherwise reject (1)       |
# +----------+---------------------------+-------+----------------------------+
    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az_to_newaz_and_host(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az('az1')
        mock_get_by_instance_uuid.return_value = fake_spec

        self.compute_api.unshelve(
            context, instance, new_az=fake_zone, host=fake_host
        )

        self._assert_unshelving_and_request_spec_az_and_host(
            context,
            instance,
            fake_spec,
            fake_zone,
            fake_host,
            mock_get_by_instance_uuid,
            mock_unshelve
        )

    @mock.patch.object(nova.conductor.ComputeTaskAPI, 'unshelve_instance')
    @mock.patch.object(objects.RequestSpec, 'save')
    @mock.patch.object(objects.ComputeNodeList, 'get_all_by_host')
    @mock.patch.object(objects.RequestSpec, 'get_by_instance_uuid')
    def test_unshelve_with_az_to_newaz_and_invalid_host(
            self,
            mock_get_by_instance_uuid,
            mock_get_all_by_host,
            mock_save,
            mock_unshelve
        ):

        context = self.context.elevated()
        fake_host = 'fake_host1'
        fake_zone = 'avail_zone1'
        self._create_host_inside_az(self.context, fake_host, fake_zone)

        instance = self._get_specify_state_instance(
            vm_states.SHELVED_OFFLOADED)

        fake_spec = self._create_request_spec_for_initial_az('az1')
        mock_get_by_instance_uuid.return_value = fake_spec

        exc = self.assertRaises(
            nova.exception.UnshelveHostNotInAZ,
            self.compute_api.unshelve,
            context,
            instance,
            new_az=fake_zone,
            host='fake_mini'
        )

        self.assertIn(
            exc.message,
            'Host "fake_mini" is not in the availability zone "avail_zone1".'
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils.fixture import uuidsentinel as uuids

from nova.compute import api as compute_api
from nova import context as nova_context
from nova import exception
from nova import objects
from nova import test
from nova.tests import fixtures as nova_fixtures


class ComputeAPITestCase(test.NoDBTestCase):
    USES_DB_SELF = True

    def setUp(self):
        super(ComputeAPITestCase, self).setUp()
        self.useFixture(nova_fixtures.Database(database='api'))

    @mock.patch('nova.objects.instance_mapping.InstanceMapping.create')
    def test_reqspec_buildreq_instmapping_single_transaction(self,
                                                             mock_create):
        # Simulate a DBError during an INSERT by raising an exception from the
        # InstanceMapping.create method.
        mock_create.side_effect = test.TestingException('oops')

        ctxt = nova_context.RequestContext('fake-user', 'fake-project')
        rs = objects.RequestSpec(context=ctxt, instance_uuid=uuids.inst)
        # project_id and instance cannot be None
        br = objects.BuildRequest(context=ctxt, instance_uuid=uuids.inst,
                                  project_id=ctxt.project_id,
                                  instance=objects.Instance())
        im = objects.InstanceMapping(context=ctxt, instance_uuid=uuids.inst)

        self.assertRaises(
            test.TestingException,
            compute_api.API._create_reqspec_buildreq_instmapping, ctxt, rs, br,
            im)

        # Since the instance mapping failed to INSERT, we should not have
        # written a request spec record or a build request record.
        self.assertRaises(
            exception.RequestSpecNotFound,
            objects.RequestSpec.get_by_instance_uuid, ctxt, uuids.inst)
        self.assertRaises(
            exception.BuildRequestNotFound,
            objects.BuildRequest.get_by_instance_uuid, ctxt, uuids.inst)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Fixtures solely for functional tests."""
import fixtures
from keystoneauth1 import adapter as ka
from keystoneauth1 import session as ks
from placement.tests.functional.fixtures import placement as placement_fixtures
from requests import adapters

from nova.tests.functional.api import client


class PlacementApiClient(object):
    def __init__(self, placement_fixture):
        self.fixture = placement_fixture

    def get(self, url, **kwargs):
        return client.APIResponse(
            self.fixture._fake_get(None, url, **kwargs))

    def put(self, url, body, **kwargs):
        return client.APIResponse(
            self.fixture._fake_put(None, url, body, **kwargs))

    def post(self, url, body, **kwargs):
        return client.APIResponse(
            self.fixture._fake_post(None, url, body, **kwargs))

    def delete(self, url, **kwargs):
        return client.APIResponse(
            self.fixture._fake_delete(None, url, **kwargs))


class PlacementFixture(placement_fixtures.PlacementFixture):
    """A fixture to placement operations.

    Runs a local WSGI server bound on a free port and having the Placement
    application with NoAuth middleware.
    This fixture also prevents calling the ServiceCatalog for getting the
    endpoint.

    It's possible to ask for a specific token when running the fixtures so
    all calls would be passing this token.

    Most of the time users of this fixture will also want the placement
    database fixture to be called first, so that is done automatically. If
    that is not desired pass ``db=False`` when initializing the fixture
    and establish the database yourself with:

        self.useFixture(placement_fixtures.Database(set_config=True))
    """

    def setUp(self):
        super(PlacementFixture, self).setUp()

        # Turn off manipulation of socket_options in TCPKeepAliveAdapter
        # to keep wsgi-intercept happy. Replace it with the method
        # from its superclass.
        self.useFixture(fixtures.MonkeyPatch(
            'keystoneauth1.session.TCPKeepAliveAdapter.init_poolmanager',
            adapters.HTTPAdapter.init_poolmanager))

        self._client = ka.Adapter(ks.Session(auth=None), raise_exc=False)
        # NOTE(sbauza): We need to mock the scheduler report client because
        # we need to fake Keystone by directly calling the endpoint instead
        # of looking up the service catalog, like we did for the OSAPIFixture.
        self.useFixture(fixtures.MonkeyPatch(
            'nova.scheduler.client.report.SchedulerReportClient.get',
            self._fake_get))
        self.useFixture(fixtures.MonkeyPatch(
            'nova.scheduler.client.report.SchedulerReportClient.post',
            self._fake_post))
        self.useFixture(fixtures.MonkeyPatch(
            'nova.scheduler.client.report.SchedulerReportClient.put',
            self._fake_put))
        self.useFixture(fixtures.MonkeyPatch(
            'nova.scheduler.client.report.SchedulerReportClient.delete',
            self._fake_delete))

        self.api = PlacementApiClient(self)

    @staticmethod
    def _update_headers_with_version(headers, version):
        if version is not None:
            # TODO(mriedem): Perform some version discovery at some point.
            headers.update({
                'OpenStack-API-Version': 'placement %s' % version
            })

    def _fake_get(self, client, url, version=None, global_request_id=None):
        # TODO(sbauza): The current placement NoAuthMiddleware returns a 401
        # in case a token is not provided. We should change that by creating
        # a fake token so we could remove adding the header below.
        headers = {'x-auth-token': self.token}
        self._update_headers_with_version(headers, version)
        return self._client.get(
            url,
            endpoint_override=self.endpoint,
            headers=headers)

    def _fake_post(
        self, client, url, data, version=None, global_request_id=None
    ):
        # NOTE(sdague): using json= instead of data= sets the
        # media type to application/json for us. Placement API is
        # more sensitive to this than other APIs in the OpenStack
        # ecosystem.
        # TODO(sbauza): The current placement NoAuthMiddleware returns a 401
        # in case a token is not provided. We should change that by creating
        # a fake token so we could remove adding the header below.
        headers = {'x-auth-token': self.token}
        self._update_headers_with_version(headers, version)
        return self._client.post(
            url, json=data,
            endpoint_override=self.endpoint,
            headers=headers)

    def _fake_put(
        self, client, url, data, version=None, global_request_id=None
    ):
        # NOTE(sdague): using json= instead of data= sets the
        # media type to application/json for us. Placement API is
        # more sensitive to this than other APIs in the OpenStack
        # ecosystem.
        # TODO(sbauza): The current placement NoAuthMiddleware returns a 401
        # in case a token is not provided. We should change that by creating
        # a fake token so we could remove adding the header below.
        headers = {'x-auth-token': self.token}
        self._update_headers_with_version(headers, version)
        return self._client.put(
            url, json=data,
            endpoint_override=self.endpoint,
            headers=headers)

    def _fake_delete(
        self, client, url, version=None, global_request_id=None
    ):
        # TODO(sbauza): The current placement NoAuthMiddleware returns a 401
        # in case a token is not provided. We should change that by creating
        # a fake token so we could remove adding the header below.
        headers = {'x-auth-token': self.token}
        self._update_headers_with_version(headers, version)
        return self._client.delete(
            url,
            endpoint_override=self.endpoint,
            headers=headers)



# Copyright 2012 Nebula, Inc.
# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.tests.functional.api_sample_tests import test_servers


class DeferredDeleteSampleJsonTests(test_servers.ServersSampleBase):
    sample_dir = "os-deferred-delete"

    def setUp(self):
        super(DeferredDeleteSampleJsonTests, self).setUp()
        self.flags(reclaim_instance_interval=1)

    def test_restore(self):
        uuid = self._post_server()
        self._do_delete('servers/%s' % uuid)

        response = self._do_post('servers/%s/action' % uuid,
                                 'restore-post-req', {})
        self.assertEqual(202, response.status_code)
        self.assertEqual('', response.text)

    def test_force_delete(self):
        uuid = self._post_server()
        self._do_delete('servers/%s' % uuid)

        response = self._do_post('servers/%s/action' % uuid,
                                 'force-delete-post-req', {})
        self.assertEqual(202, response.status_code)
        self.assertEqual('', response.text)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
# Copyright (c) 2010 Citrix Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Power state is the state we get by calling virt driver on a particular
domain. The hypervisor is always considered the authority on the status
of a particular VM, and the power_state in the DB should be viewed as a
snapshot of the VMs's state in the (recent) past. It can be periodically
updated, and should also be updated at the end of a task if the task is
supposed to affect power_state.
"""

from nova.objects import fields

# NOTE(maoy): These are *not* virDomainState values from libvirt.
# The hex value happens to match virDomainState for backward-compatibility
# reasons.
NOSTATE = fields.InstancePowerState.index(fields.InstancePowerState.NOSTATE)
RUNNING = fields.InstancePowerState.index(fields.InstancePowerState.RUNNING)
PAUSED = fields.InstancePowerState.index(fields.InstancePowerState.PAUSED)
# the VM is powered off
SHUTDOWN = fields.InstancePowerState.index(fields.InstancePowerState.SHUTDOWN)
CRASHED = fields.InstancePowerState.index(fields.InstancePowerState.CRASHED)
SUSPENDED = fields.InstancePowerState.index(
    fields.InstancePowerState.SUSPENDED)

# TODO(justinsb): Power state really needs to be a proper class,
# so that we're not locked into the libvirt status codes and can put mapping
# logic here rather than spread throughout the code
STATE_MAP = {fields.InstancePowerState.index(state): state
             for state in fields.InstancePowerState.ALL
             if state != fields.InstancePowerState._UNUSED}



# Copyright (C) 2016-2020 Lightbits Labs Ltd.
# Copyright (C) 2020 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import nova.conf
from nova import utils
from nova.virt.libvirt.volume import volume as libvirt_volume
from os_brick import initiator
from os_brick.initiator import connector
from oslo_log import log as logging


LOG = logging.getLogger(__name__)
CONF = nova.conf.CONF


class LibvirtLightOSVolumeDriver(libvirt_volume.LibvirtVolumeDriver):
    """Driver to attach NVMe volumes to libvirt."""
    VERSION = '2.3.12'

    def __init__(self, connection):
        super(LibvirtLightOSVolumeDriver, self).__init__(connection)
        self.connector = connector.InitiatorConnector.factory(
            initiator.LIGHTOS,
            root_helper=utils.get_root_helper(),
            device_scan_attempts=CONF.libvirt.num_nvme_discover_tries)

    def connect_volume(self, connection_info, instance):
        device_info = self.connector.connect_volume(connection_info['data'])
        LOG.debug("Connecting NVMe volume with device_info %s", device_info)
        connection_info['data']['device_path'] = device_info['path']

    def disconnect_volume(self, connection_info, instance, force=False):
        """Detach the volume from the instance."""
        LOG.debug("Disconnecting NVMe disk. instance:%s, volume_id:%s",
                  connection_info.get("instance", ""),
                  connection_info.get("volume_id", ""))
        self.connector.disconnect_volume(
            connection_info['data'], None, force=force)
        super(LibvirtLightOSVolumeDriver, self).disconnect_volume(
            connection_info, instance, force=force)

    def extend_volume(self, connection_info, instance, requested_size):
        """Extend the volume."""
        LOG.debug("calling os-brick to extend LightOS Volume."
                  "instance:%s, volume_id:%s",
                  connection_info.get("instance", ""),
                  connection_info.get("volume_id", ""))
        new_size = self.connector.extend_volume(connection_info['data'])
        LOG.debug("Extend LightOS Volume %s; new_size=%s",
                  connection_info['data']['device_path'], new_size)
        return new_size



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_utils import uuidutils


from nova.tests.fixtures import libvirt as fakelibvirt
from nova.tests.functional.libvirt import test_vgpu
from nova.virt.libvirt import utils as libvirt_utils


class VGPUTestsLibvirt7_7(test_vgpu.VGPUTestBase):

    def _create_mdev(self, physical_device, mdev_type, uuid=None):
        # We need to fake the newly created sysfs object by adding a new
        # FakeMdevDevice in the existing persisted Connection object so
        # when asking to get the existing mdevs, we would see it.
        if not uuid:
            uuid = uuidutils.generate_uuid()
        mdev_name = libvirt_utils.mdev_uuid2name(uuid)
        libvirt_parent = self.pci2libvirt_address(physical_device)

        # Libvirt 7.7 now creates mdevs with a parent_addr suffix.
        new_mdev_name = '_'.join([mdev_name, libvirt_parent])

        # Here, we get the right compute thanks by the self.current_host that
        # was modified just before
        connection = self.computes[
            self._current_host].driver._host.get_connection()
        connection.mdev_info.devices.update(
            {mdev_name: fakelibvirt.FakeMdevDevice(dev_name=new_mdev_name,
                                                   type_id=mdev_type,
                                                   parent=libvirt_parent)})
        return uuid

    def setUp(self):
        super(VGPUTestsLibvirt7_7, self).setUp()
        extra_spec = {"resources:VGPU": "1"}
        self.flavor = self._create_flavor(extra_spec=extra_spec)

        # Start compute1 supporting only nvidia-11
        self.flags(
            enabled_mdev_types=fakelibvirt.NVIDIA_11_VGPU_TYPE,
            group='devices')

        self.compute1 = self.start_compute_with_vgpu('host1')

    def test_create_servers_with_vgpu(self):

        # Create a single instance against a specific compute node.
        self._create_server(
            image_uuid='155d900f-4e14-4e4c-a73d-069cbf4541e6',
            flavor_id=self.flavor, host=self.compute1.host,
            networks='auto', expected_state='ACTIVE')

        self.assert_mdev_usage(self.compute1, expected_amount=1)

        self._create_server(
            image_uuid='155d900f-4e14-4e4c-a73d-069cbf4541e6',
            flavor_id=self.flavor, host=self.compute1.host,
            networks='auto', expected_state='ACTIVE')

        self.assert_mdev_usage(self.compute1, expected_amount=2)



# Copyright (c) 2012 VMware, Inc.
# Copyright (c) 2011 Citrix Systems, Inc.
# Copyright 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Utility functions for ESX Networking.
"""
import re

from oslo_log import log as logging
from oslo_vmware import vim_util as vutil

from nova.virt.vmwareapi import vim_util
from nova.virt.vmwareapi import vm_util

LOG = logging.getLogger(__name__)

# a virtual wire will have the following format:
# vxw-<dvs-moref>-<virtualwire-moref>-<sid-moref>-<name>
# Examples:
# - vxw-dvs-22-virtualwire-89-sid-5008-NAME
# - vxw-dvs-22-virtualwire-89-sid-5008-UUID
VWIRE_REGEX = re.compile(r'vxw-dvs-(\d+)-virtualwire-(\d+)-sid-(\d+)-(.*)')


def _get_name_from_dvs_name(dvs_name):
    vwire = VWIRE_REGEX.match(dvs_name)
    if not vwire:
        return dvs_name
    return vwire.group(4)


def _get_network_obj(session, network_objects, network_name):
    """Gets the network object for the requested network.

    The network object will be used when creating the VM configuration
    spec. The network object contains the relevant network details for
    the specific network type, for example, a distributed port group.

    The method will search for the network_name in the list of
    network_objects.

    :param session: vCenter soap session
    :param network_objects: group of networks
    :param network_name: the requested network
    :return: network object
    """

    network_obj = {}
    # network_objects is actually a RetrieveResult object from vSphere API call
    for obj_content in network_objects:
        # the propset attribute "need not be set" by returning API
        if not hasattr(obj_content, 'propSet'):
            continue
        prop_dict = vm_util.propset_dict(obj_content.propSet)
        network_refs = prop_dict.get('network')
        if network_refs:
            network_refs = network_refs.ManagedObjectReference
            for network in network_refs:
                # Get network properties
                if network._type == 'DistributedVirtualPortgroup':
                    props = session._call_method(vutil,
                                                 "get_object_property",
                                                 network,
                                                 "config")
                    # NOTE(asomya): This only works on ESXi if the port binding
                    # is set to ephemeral
                    net_name = _get_name_from_dvs_name(props.name)
                    if network_name in net_name:
                        network_obj['type'] = 'DistributedVirtualPortgroup'
                        network_obj['dvpg'] = props.key
                        dvs_props = session._call_method(vutil,
                                        "get_object_property",
                                        props.distributedVirtualSwitch,
                                        "uuid")
                        network_obj['dvsw'] = dvs_props
                        return network_obj
                else:
                    props = session._call_method(vutil,
                                                 "get_object_property",
                                                 network,
                                                 "summary.name")
                    if props == network_name:
                        network_obj['type'] = 'Network'
                        network_obj['name'] = network_name
                        return network_obj


def get_network_with_the_name(session, network_name="vmnet0", cluster=None):
    """Gets reference to the network whose name is passed as the
    argument.
    """
    vm_networks = session._call_method(vim_util,
                                       'get_object_properties',
                                       None, cluster,
                                       'ClusterComputeResource', ['network'])

    with vutil.WithRetrieval(session.vim, vm_networks) as network_objs:
        network_obj = _get_network_obj(session, network_objs, network_name)
        if network_obj:
            return network_obj

    LOG.debug("Network %s not found on cluster!", network_name)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures
import webob

from nova.api.openstack.compute import multinic as multinic_v21
from nova import compute
from nova import exception
from nova import objects
from nova import test
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_instance


UUID = '70f6db34-de8d-4fbd-aafb-4065bdfa6114'
last_add_fixed_ip = (None, None)
last_remove_fixed_ip = (None, None)


def compute_api_add_fixed_ip(self, context, instance, network_id):
    global last_add_fixed_ip

    last_add_fixed_ip = (instance['uuid'], network_id)


def compute_api_remove_fixed_ip(self, context, instance, address):
    global last_remove_fixed_ip

    last_remove_fixed_ip = (instance['uuid'], address)


def compute_api_get(self, context, instance_id, expected_attrs=None,
                    cell_down_support=False):
    instance = objects.Instance()
    instance.uuid = instance_id
    instance.id = 1
    instance.vm_state = 'fake'
    instance.task_state = 'fake'
    instance.obj_reset_changes()
    return instance


class FixedIpTestV21(test.NoDBTestCase):
    controller_class = multinic_v21
    validation_error = exception.ValidationError

    def setUp(self):
        super(FixedIpTestV21, self).setUp()
        fakes.stub_out_networking(self)
        self.stub_out('nova.compute.api.API.add_fixed_ip',
                      compute_api_add_fixed_ip)
        self.stub_out('nova.compute.api.API.remove_fixed_ip',
                      compute_api_remove_fixed_ip)
        self.stub_out('nova.compute.api.API.get', compute_api_get)
        self.controller = self.controller_class.MultinicController()
        self.fake_req = fakes.HTTPRequest.blank('')
        self.mock_get = self.useFixture(
            fixtures.MockPatch('nova.api.openstack.common.get_instance')).mock
        self.mock_get.return_value = fake_instance.fake_instance_obj(
            self.fake_req.environ['nova.context'], uuid=UUID,
            project_id=self.fake_req.environ['nova.context'].project_id)

    def test_add_fixed_ip(self):
        global last_add_fixed_ip
        last_add_fixed_ip = (None, None)

        body = dict(addFixedIp=dict(networkId='test_net'))
        self.controller._add_fixed_ip(self.fake_req, UUID, body=body)
        self.assertEqual(last_add_fixed_ip, (UUID, 'test_net'))

    def _test_add_fixed_ip_bad_request(self, body):
        self.assertRaises(self.validation_error,
                          self.controller._add_fixed_ip,
                          self.fake_req,
                          UUID, body=body)

    def test_add_fixed_ip_empty_network_id(self):
        body = {'addFixedIp': {'network_id': ''}}
        self._test_add_fixed_ip_bad_request(body)

    def test_add_fixed_ip_network_id_bigger_than_36(self):
        body = {'addFixedIp': {'network_id': 'a' * 37}}
        self._test_add_fixed_ip_bad_request(body)

    def test_add_fixed_ip_no_network(self):
        global last_add_fixed_ip
        last_add_fixed_ip = (None, None)

        body = dict(addFixedIp=dict())
        self._test_add_fixed_ip_bad_request(body)
        self.assertEqual(last_add_fixed_ip, (None, None))

    @mock.patch.object(compute.api.API, 'add_fixed_ip')
    def test_add_fixed_ip_no_more_ips_available(self, mock_add_fixed_ip):
        mock_add_fixed_ip.side_effect = exception.NoMoreFixedIps(net='netid')

        body = dict(addFixedIp=dict(networkId='test_net'))
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._add_fixed_ip,
                          self.fake_req,
                          UUID, body=body)

    def test_remove_fixed_ip(self):
        global last_remove_fixed_ip
        last_remove_fixed_ip = (None, None)

        body = dict(removeFixedIp=dict(address='10.10.10.1'))
        self.controller._remove_fixed_ip(self.fake_req, UUID, body=body)
        self.assertEqual(last_remove_fixed_ip, (UUID, '10.10.10.1'))

    def test_remove_fixed_ip_no_address(self):
        global last_remove_fixed_ip
        last_remove_fixed_ip = (None, None)

        body = dict(removeFixedIp=dict())
        self.assertRaises(self.validation_error,
                          self.controller._remove_fixed_ip,
                          self.fake_req,
                          UUID, body=body)
        self.assertEqual(last_remove_fixed_ip, (None, None))

    def test_remove_fixed_ip_invalid_address(self):
        body = {'removeFixedIp': {'address': ''}}
        self.assertRaises(self.validation_error,
                          self.controller._remove_fixed_ip,
                          self.fake_req,
                          UUID, body=body)

    @mock.patch.object(compute.api.API, 'remove_fixed_ip',
        side_effect=exception.FixedIpNotFoundForInstance(
            instance_uuid=UUID, ip='10.10.10.1'))
    def test_remove_fixed_ip_not_found(self, _remove_fixed_ip):

        body = {'removeFixedIp': {'address': '10.10.10.1'}}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._remove_fixed_ip,
                          self.fake_req,
                          UUID, body=body)


class MultinicAPIDeprecationTest(test.NoDBTestCase):

    def setUp(self):
        super(MultinicAPIDeprecationTest, self).setUp()
        self.controller = multinic_v21.MultinicController()
        self.req = fakes.HTTPRequest.blank('', version='2.44')

    def test_add_fixed_ip_not_found(self):
        body = dict(addFixedIp=dict(networkId='test_net'))
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
            self.controller._add_fixed_ip, self.req, UUID, body=body)

    def test_remove_fixed_ip__not_found(self):
        body = dict(removeFixedIp=dict(address='10.10.10.1'))
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
            self.controller._remove_fixed_ip, self.req, UUID, body=body)



# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Websocket proxy that is compatible with OpenStack Nova
noVNC consoles. Leverages websockify.py by Joel Martin
"""

import sys


from nova.cmd import baseproxy
import nova.conf
from nova.conf import remote_debug
from nova.conf import vnc
from nova import config
from nova.console.securityproxy import rfb


CONF = nova.conf.CONF
remote_debug.register_cli_opts(CONF)
vnc.register_cli_opts(CONF)


def main():
    # set default web flag option
    CONF.set_default('web', '/usr/share/novnc')
    config.parse_args(sys.argv)

    baseproxy.proxy(
        host=CONF.vnc.novncproxy_host,
        port=CONF.vnc.novncproxy_port,
        security_proxy=rfb.RFBSecurityProxy())
